"""
An objective function indirectly minimizing the number of bins in packings.

This objective function first computes the number of bins used. Let's call it
`n_bins`. We know the total number of items,
:attr:`~moptipyapps.binpacking2d.instance.Instance.n_items`, as
well (because this is also the number of rows in the packing).
Now we return `(n_items * (n_bins - 1)) + number_of_items_in_last_bin`,
where `number_of_items_in_last_bin` is, well, the number of items in the
very last bin.

The idea behind this is: If one of two packings has the smaller number of
bins, then this one will always have the smaller objective value. If two
packings have the same number of bins, but one has fewer items in the very
last bin, then that one is better. With this mechanism, we drive the search
towards "emptying" the last bin. If the number of items in the last bin would
reach `0`, that last bin would disappear - and we have one bin less.
"""
from typing import Final

import numba  # type: ignore
import numpy as np
from moptipy.api.objective import Objective
from moptipy.utils.types import type_error

from moptipyapps.binpacking2d.instance import Instance
from moptipyapps.binpacking2d.packing import IDX_BIN


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def bin_count_and_last_empty(y: np.ndarray) -> int:
    """
    Compute the number of bins and the emptiness of the last bin.

    We compute the total number of bins minus 1 and multiply it with the
    number of items. We then add the number of items in the last bin.

    :param y: the packing
    :return: the objective value

    >>> bin_count_and_last_empty(np.array([[1, 1, 10, 10, 20, 20],
    ...                                    [1, 1, 30, 30, 40, 40],
    ...                                    [1, 1, 20, 20, 30, 30]], int))
    3
    >>> bin_count_and_last_empty(np.array([[1, 1, 10, 10, 20, 20],
    ...                                    [1, 2, 30, 30, 40, 40],  # bin 2!
    ...                                    [1, 1, 20, 20, 30, 30]], int))
    4
    >>> bin_count_and_last_empty(np.array([[1, 2, 10, 10, 20, 20],  # bin 2!
    ...                                    [1, 2, 30, 30, 40, 40],  # bin 2!
    ...                                    [1, 1, 20, 20, 30, 30]], int))
    5
    >>> bin_count_and_last_empty(np.array([[1, 3, 10, 10, 20, 20],  # bin 3!
    ...                                    [1, 2, 30, 30, 40, 40],  # bin 2!
    ...                                    [1, 1, 20, 20, 30, 30]], int))
    7
    """
    current_bin: int = -1  # the current idea of what the last bin is
    current_size: int = -1  # the number of items already in that bin
    n_items: Final[int] = len(y)  # the number of rows in the matrix

    for i in range(n_items):  # iterate over all packed items
        bin_idx: int = int(y[i, IDX_BIN])  # get the bin index of the item
        if bin_idx > current_bin:  # it's a new biggest bin = new last bin?
            current_size = 1  # then there is 1 object in it for now
            current_bin = bin_idx  # and we remember it
        elif bin_idx == current_bin:  # did item go into the current last bin?
            current_size = current_size + 1  # then increase size
    return (n_items * (current_bin - 1)) + current_size  # return objective


class BinCountAndLastEmpty(Objective):
    """Compute the number of bins and the emptiness of the last one."""

    def __init__(self, instance: Instance) -> None:
        """
        Initialize the objective function.

        :param instance: the instance to load the bounds from
        """
        super().__init__()
        if not isinstance(instance, Instance):
            raise type_error(instance, "instance", Instance)
        #: the internal instance reference
        self.__instance: Final[Instance] = instance

    def evaluate(self, x) -> int:
        """
        Evaluate the objective function.

        :param x: the solution
        :return: the bin size and emptyness factor
        """
        return bin_count_and_last_empty(x)

    def lower_bound(self) -> int:
        """
        Get the lower bound of the number of bins and emptiness objective.

        We know from the instance (:attr:`~moptipyapps.binpacking2d\
.instance.Instance.lower_bound_bins`) that we require at least as many bins
        such that they can accommodate the total area of all items together.
        Let's call this number `lb`. Now if `lb` is one, then all objects could
        be in the first bin, in which case the objective value would equal to
        :attr:`~moptipyapps.binpacking2d.instance.Instance.n_items`,
        i.e., the total number of items in the first = last bin.
        If it is `lb=2`, then we know that we will need at least two bins. The
        best case would be that `n_items - 1` items are in the first bin and
        one is in the last bin. This means that we would get `1 * n_items + 1`
        as objective value. If we have `lb=3` bins, then we could have
        `n_items - 1` items distributed over the first two bins with one item
        left over in the last bin, i.e., would get `(2 * n_items) + 1`. And so
        on.

        :return: `max(n_items, (lb - 1) * n_items + 1)`

        >>> ins = Instance("a", 100, 50, [[10, 5, 1], [3, 3, 1], [5, 5, 1]])
        >>> ins.n_items
        3
        >>> ins.lower_bound_bins
        1
        >>> BinCountAndLastEmpty(ins).lower_bound()
        3

        >>> ins = Instance("b", 10, 50, [[10, 5, 10], [3, 3, 1], [5, 5, 1]])
        >>> ins.n_items
        12
        >>> ins.lower_bound_bins
        2
        >>> BinCountAndLastEmpty(ins).lower_bound()
        13

        >>> ins = Instance("c", 10, 50, [[10, 5, 20], [30, 3, 10], [5, 5, 1]])
        >>> ins.n_items
        31
        >>> ins.lower_bound_bins
        4
        >>> BinCountAndLastEmpty(ins).lower_bound()
        94
        """
        return max(self.__instance.n_items,
                   ((self.__instance.lower_bound_bins - 1)
                    * self.__instance.n_items) + 1)

    def is_always_integer(self) -> bool:
        """
        Return `True` because there are only integer bins.

        :retval True: always
        """
        return True

    def upper_bound(self) -> int:
        """
        Get the upper bound of the number of bins plus emptiness.

        :return: the number of items in the instance to the square

        >>> ins = Instance("a", 100, 50, [[10, 5, 1], [3, 3, 1], [5, 5, 1]])
        >>> ins.n_items
        3
        >>> BinCountAndLastEmpty(ins).upper_bound()
        9

        >>> ins = Instance("b", 10, 50, [[10, 5, 10], [3, 3, 1], [5, 5, 1]])
        >>> ins.n_items
        12
        >>> BinCountAndLastEmpty(ins).upper_bound()
        144

        >>> ins = Instance("c", 10, 50, [[10, 5, 20], [30, 3, 10], [5, 5, 1]])
        >>> ins.n_items
        31
        >>> BinCountAndLastEmpty(ins).upper_bound()
        961
        """
        return self.__instance.n_items * self.__instance.n_items

    def __str__(self) -> str:
        """
        Get the name of the bins objective function.

        :return: `binCountAndLastEmpty`
        :retval "binCountAndLastEmpty": always
        """
        return "binCountAndLastEmpty"



"""
An extension of the permutation space for one-dimensional ordering.

The main difference is how the `to_str` result is noted, namely that it
contains the mapping of tags to locations.

>>> def _dist(a, b):
...     return abs(a - b)
>>> def _tags(a):
...     return f"t{a}"
>>> the_instance = Instance.from_sequence_and_distance(
...     [1, 2, 3, 3, 2, 3], _dist, 2, 10, ("x", ), _tags)
>>> the_space = OrderingSpace(the_instance)
>>> the_str = the_space.to_str(np.array([0, 2, 1]))
>>> the_str.splitlines()
['0;2;1', '', 'indexZeroBased;suggestedXin01;x', '0;0;t1', '2;1;t2', \
'2;1;t2', '1;0.5;t3', '1;0.5;t3', '1;0.5;t3']
>>> print(the_space.from_str(the_str))
[0 2 1]
"""

from typing import Final

import numpy as np
from moptipy.spaces.permutations import Permutations
from moptipy.utils.logger import CSV_SEPARATOR
from moptipy.utils.strings import float_to_str
from moptipy.utils.types import type_error

from moptipyapps.order1d.instance import (
    _SUGGESTED_X_IN_0_1,
    _ZERO_BASED_INDEX,
    Instance,
)


class OrderingSpace(Permutations):
    """A space for one-dimensional orderings."""

    def __init__(self, instance: Instance) -> None:
        """
        Create an ordering space from an instance.

        :param instance: the instance
        """
        if not isinstance(instance, Instance):
            raise type_error(instance, "instance", Instance)
        super().__init__(range(instance.n))
        #: the instance
        self.instance: Final[Instance] = instance

    def to_str(self, x: np.ndarray) -> str:
        """
        Convert a solution to a string.

        :param x: the permutation
        :return: the string
        """
        tags: Final[tuple[tuple[tuple[str, ...], int], ...]] \
            = self.instance.tags

        n: Final[int] = len(self.blueprint) - 1
        text: list[str] = []

        text.extend(super().to_str(x).split("\n"))
        text.append("")

        row: list[str] = [_ZERO_BASED_INDEX, _SUGGESTED_X_IN_0_1]
        row.extend(self.instance.tag_titles)
        text.append(f"{CSV_SEPARATOR.join(row)}")

        for tag, i in tags:
            row.clear()
            row.append(str(x[i]))
            row.append(float_to_str(x[i] / n))
            row.extend(tag)
            text.append(f"{CSV_SEPARATOR.join(row)}")
        return "\n".join(text)

    def from_str(self, text: str) -> np.ndarray:
        """
        Get the string version from the given text.

        :param text: the text
        :return: the string
        """
        text = text.lstrip()
        idx: int = text.find("\n")
        if idx > 0:
            text = text[:idx]
        return super().from_str(text.rstrip())



"""Test all the example code in the project's examples directory."""
import os.path
from typing import Final

from moptipy.utils.console import logger
from moptipy.utils.path import Path
from moptipy.utils.temp import TempDir
from numpy.random import default_rng


def test_examples_in_examples_directory() -> None:
    """Test all the examples in the examples directory."""
    # First, we resolve the directories
    base_dir = Path.directory(os.path.join(os.path.dirname(__file__), "../"))
    examples_dir = Path.directory(base_dir.resolve_inside("examples"))
    logger(
        f"executing all examples from examples directory {examples_dir!r}.")

    wd: Final[str] = os.getcwd()  # get current working directory
    with TempDir.create() as td:  # create temporary working directory
        logger(f"using temp directory {td!r}.")
        os.chdir(td)  # set it as working directory
        files: list[str] = os.listdir(examples_dir)
        logger(f"got {len(files)} potential files")
        default_rng().shuffle(files)  # shuffle the order for randomness
        for name in files:  # find all files in examples
            if name.endswith(".py"):  # if it is a python file
                file: Path = Path.file(examples_dir.resolve_inside(name))
                logger(f"now compiling file {file!r}.")
                code = compile(  # noqa # nosec
                    file.read_all_str(), file, mode="exec")  # noqa # nosec
                logger(f"now executing file {file!r}.")
                exec(code, {})  # execute file  # noqa # nosec
                logger(f"successfully executed example {file!r}.")
    os.chdir(wd)  # go back to current original directory
    logger("finished executing all examples from README.md.")



"""
An extended end result record to represent packings.

This class extends the information provided by
:mod:`~moptipy.evaluation.end_results`. It allows us to compare the results
of experiments over different objective functions. It also represents the
bounds for the number of bins and for the objective functions. It also
includes the problem-specific information of two-dimensional bin packing
instances.
"""
import argparse
import os.path
from dataclasses import dataclass
from math import isfinite
from typing import Any, Callable, Final, Iterable, Mapping, cast

from moptipy.api.logging import (
    KEY_ALGORITHM,
    KEY_BEST_F,
    KEY_GOAL_F,
    KEY_INSTANCE,
    KEY_LAST_IMPROVEMENT_FE,
    KEY_LAST_IMPROVEMENT_TIME_MILLIS,
    KEY_MAX_FES,
    KEY_MAX_TIME_MILLIS,
    KEY_RAND_SEED,
    KEY_TOTAL_FES,
    KEY_TOTAL_TIME_MILLIS,
)
from moptipy.api.objective import Objective
from moptipy.evaluation.base import (
    KEY_ENCODING,
    KEY_OBJECTIVE_FUNCTION,
    EvaluationDataElement,
)
from moptipy.evaluation.end_results import EndResult
from moptipy.evaluation.log_parser import LogParser
from moptipy.utils.console import logger
from moptipy.utils.help import argparser
from moptipy.utils.logger import CSV_SEPARATOR
from moptipy.utils.path import Path
from moptipy.utils.strings import (
    intfloatnone_to_str,
    intnone_to_str,
    num_to_str,
    str_to_intfloat,
    str_to_intfloatnone,
    str_to_intnone,
)
from moptipy.utils.types import check_int_range, immutable_mapping, type_error

from moptipyapps.binpacking2d.instance import Instance, _lower_bound_damv
from moptipyapps.binpacking2d.objectives.bin_count import (
    BIN_COUNT_NAME,
    BinCount,
)
from moptipyapps.binpacking2d.objectives.bin_count_and_empty import (
    BinCountAndEmpty,
)
from moptipyapps.binpacking2d.objectives.bin_count_and_last_empty import (
    BinCountAndLastEmpty,
)
from moptipyapps.binpacking2d.objectives.bin_count_and_last_small import (
    BinCountAndLastSmall,
)
from moptipyapps.binpacking2d.objectives.bin_count_and_lowest_skyline import (
    BinCountAndLowestSkyline,
)
from moptipyapps.binpacking2d.objectives.bin_count_and_small import (
    BinCountAndSmall,
)
from moptipyapps.binpacking2d.packing import Packing

#: the number of items
KEY_N_ITEMS: Final[str] = "nItems"
#: the number of different items
KEY_N_DIFFERENT_ITEMS: Final[str] = "nDifferentItems"
#: the bin width
KEY_BIN_WIDTH: Final[str] = "binWidth"
#: the bin height
KEY_BIN_HEIGHT: Final[str] = "binHeight"

#: The internal CSV header part 1, after which the encoding may come
_HEADER_1: Final[str] = (f"{KEY_ALGORITHM}{CSV_SEPARATOR}"
                         f"{KEY_INSTANCE}{CSV_SEPARATOR}"
                         f"{KEY_OBJECTIVE_FUNCTION}")
#: The internal CSV header part 2, after which the bounds are inserted
_HEADER_2: Final[str] = (f"{KEY_N_ITEMS}{CSV_SEPARATOR}"
                         f"{KEY_N_DIFFERENT_ITEMS}{CSV_SEPARATOR}"
                         f"{KEY_BIN_WIDTH}{CSV_SEPARATOR}"
                         f"{KEY_BIN_HEIGHT}")
#: The internal CSV header part 3, after which the objective values are listed
_HEADER_3: Final[str] = (f"{KEY_RAND_SEED}{CSV_SEPARATOR}"
                         f"{KEY_BEST_F}{CSV_SEPARATOR}"
                         f"{KEY_LAST_IMPROVEMENT_FE}{CSV_SEPARATOR}"
                         f"{KEY_LAST_IMPROVEMENT_TIME_MILLIS}"
                         f"{CSV_SEPARATOR}"
                         f"{KEY_TOTAL_FES}{CSV_SEPARATOR}"
                         f"{KEY_TOTAL_TIME_MILLIS}")


#: the default objective functions
DEFAULT_OBJECTIVES: Final[tuple[Callable[[Instance], Objective], ...]] = (
    BinCount, BinCountAndLastEmpty, BinCountAndLastSmall,
    BinCountAndLowestSkyline, BinCountAndEmpty, BinCountAndSmall,
)


def __lb_geometric(inst: Instance) -> int:
    """
    Compute the geometric lower bound.

    :param inst: the instance
    :return: the lower bound
    """
    area: Final[int] = sum(int(row[0]) * int(row[1]) * int(row[2])
                           for row in inst)
    bin_size: Final[int] = inst.bin_width * inst.bin_height
    res: int = area // bin_size
    return (res + 1) if ((res * bin_size) != area) else res


#: the lower bound of an objective
_OBJECTIVE_LOWER: Final[str] = ".lowerBound"
#: the upper bound of an objective
_OBJECTIVE_UPPER: Final[str] = ".upperBound"
#: the start string for bin bounds
_BINS_START: Final[str] = f"bins{_OBJECTIVE_LOWER}"
#: the default bounds
_DEFAULT_BOUNDS: Final[Mapping[str, Callable[[Instance], int]]] = \
    immutable_mapping({
        _BINS_START: lambda i: i.lower_bound_bins,
        f"{_BINS_START}.geometric": __lb_geometric,
        f"{_BINS_START}.damv": lambda i: _lower_bound_damv(
            i.bin_width, i.bin_height, i),
    })


@dataclass(frozen=True, init=False, order=False, eq=False)
class PackingResult(EvaluationDataElement):
    """
    An end result record of one run of one packing algorithm on one problem.

    This record provides the information of the outcome of one application of
    one algorithm to one problem instance in an immutable way.
    """

    #: the original end result record
    end_result: EndResult
    #: the number of items in the instance
    n_items: int
    #: the number of different items in the instance
    n_different_items: int
    #: the bin width
    bin_width: int
    #: the bin heigth
    bin_height: int
    #: the objective values evaluated after the optimization
    objectives: Mapping[str, int]
    #: the bounds for the objective values (append ".lowerBound" and
    #: ".upperBound" to all objective function names)
    objective_bounds: Mapping[str, int]
    #: the bounds for the minimum number of bins of the instance
    bin_bounds: Mapping[str, int]

    def __init__(self,
                 end_result: EndResult,
                 n_items: int,
                 n_different_items: int,
                 bin_width: int,
                 bin_height: int,
                 objectives: Mapping[str, int | float],
                 objective_bounds: Mapping[str, int | float],
                 bin_bounds: Mapping[str, int]):
        """
        Create a consistent instance of :class:`PackingResult`.

        :param end_result: the end result
        :param n_items: the number of items
        :param n_different_items: the number of different items
        :param bin_width: the bin width
        :param bin_height: the bin height
        :param objectives: the objective values computed after the
            optimization
        :param bin_bounds: the different bounds for the number of bins
        :param objective_bounds: the bounds for the objective functions
        :raises TypeError: if any parameter has a wrong type
        :raises ValueError: if the parameter values are inconsistent
        """
        super().__init__()
        if not isinstance(end_result, EndResult):
            raise type_error(end_result, "end_result", EndResult)
        if end_result.best_f != objectives[end_result.objective]:
            raise ValueError(
                f"end_result.best_f={end_result.best_f}, but objectives["
                f"{end_result.objective!r}]="
                f"{objectives[end_result.objective]}.")
        if not isinstance(objectives, Mapping):
            raise type_error(objectives, "objectives", Mapping)
        if not isinstance(objective_bounds, Mapping):
            raise type_error(objective_bounds, "objective_bounds", Mapping)
        if not isinstance(bin_bounds, Mapping):
            raise type_error(bin_bounds, "bin_bounds", Mapping)
        if len(objective_bounds) != (2 * len(objectives)):
            raise ValueError(f"it is required that there is a lower and an "
                             f"upper bound for each of the {len(objectives)} "
                             f"functions, but we got {len(objective_bounds)} "
                             f"bounds, objectives={objectives}, "
                             f"objective_bounds={objective_bounds}.")

        for name, value in objectives.items():
            if not isinstance(name, str):
                raise type_error(
                    name, f"name of evaluation[{name!r}]={value!r}", str)
            if not isinstance(value, int | float):
                raise type_error(
                    value, f"value of evaluation[{name!r}]={value!r}",
                    (int, float))
            if not isfinite(value):
                raise ValueError(
                    f"non-finite value of evaluation[{name!r}]={value!r}")
            lower = objective_bounds[f"{name}{_OBJECTIVE_LOWER}"]
            if not isfinite(lower):
                raise ValueError(f"{name}{_OBJECTIVE_LOWER}=={lower}.")
            upper = objective_bounds[f"{name}{_OBJECTIVE_UPPER}"]
            if not (lower <= value <= upper):
                raise ValueError(
                    f"it is required that {name}{_OBJECTIVE_LOWER}<=f<={name}"
                    f"{_OBJECTIVE_UPPER}, but got {lower}, {value}, and "
                    f"{upper}.")

        bins: Final[int | None] = cast(int, objectives[BIN_COUNT_NAME]) \
            if BIN_COUNT_NAME in objectives else None
        for name, value in bin_bounds.items():
            if not isinstance(name, str):
                raise type_error(
                    name, f"name of bounds[{name!r}]={value!r}", str)
            check_int_range(value, f"bounds[{name!r}]", 1, 1_000_000_000)
            if (bins is not None) and (bins < value):
                raise ValueError(
                    f"number of bins={bins} is inconsistent with "
                    f"bound {name!r}={value}.")

        object.__setattr__(self, "end_result", end_result)
        object.__setattr__(self, "objectives", immutable_mapping(objectives))
        object.__setattr__(self, "objective_bounds",
                           immutable_mapping(objective_bounds))
        object.__setattr__(self, "bin_bounds", immutable_mapping(bin_bounds))
        object.__setattr__(self, "n_different_items", check_int_range(
            n_different_items, "n_different_items", 1, 1_000_000_000_000))
        object.__setattr__(self, "n_items", check_int_range(
            n_items, "n_items", n_different_items, 1_000_000_000_000))
        object.__setattr__(self, "bin_width", check_int_range(
            bin_width, "bin_width", 1, 1_000_000_000_000))
        object.__setattr__(self, "bin_height", check_int_range(
            bin_height, "bin_height", 1, 1_000_000_000_000))

    def _tuple(self) -> tuple[Any, ...]:
        """
        Create a tuple with all the data of this data class for comparison.

        :returns: a tuple with all the data of this class, where `None` values
            are masked out
        """
        # noinspection PyProtectedMember
        return self.end_result._tuple()

    @staticmethod
    def from_packing_and_end_result(  # pylint: disable=W0102
            end_result: EndResult, packing: Packing,
            objectives: Iterable[Callable[[Instance], Objective]] =
            DEFAULT_OBJECTIVES,
            bin_bounds: Mapping[str, Callable[[Instance], int]] =
            _DEFAULT_BOUNDS,
            cache: Mapping[str, tuple[Mapping[str, int], tuple[
                Objective, ...], Mapping[str, int | float]]] | None =
            None) -> "PackingResult":
        """
        Create a `PackingResult` from an `EndResult` and a `Packing`.

        :param end_result: the end results record
        :param packing: the packing
        :param bin_bounds: the bounds computing functions
        :param objectives: the objective function factories
        :param cache: a cache that can store stuff if this function is to be
            called repeatedly
        :return: the packing result
        """
        if not isinstance(end_result, EndResult):
            raise type_error(end_result, "end_result", EndResult)
        if not isinstance(packing, Packing):
            raise type_error(packing, "packing", Packing)
        if not isinstance(objectives, Iterable):
            raise type_error(objectives, "objectives", Iterable)
        if not isinstance(bin_bounds, Mapping):
            raise type_error(bin_bounds, "bin_bounds", Mapping)
        if (cache is not None) and (not isinstance(cache, dict)):
            raise type_error(cache, "cache", (None, dict))

        instance: Final[Instance] = packing.instance
        if instance.name != end_result.instance:
            raise ValueError(
                f"packing.instance.name={instance.name!r}, but "
                f"end_result.instance={end_result.instance!r}.")

        row: tuple[Mapping[str, int], tuple[Objective, ...],
                   Mapping[str, int | float]] | None = None \
            if (cache is None) else cache.get(instance.name, None)
        if row is None:
            objfs = tuple(sorted((obj(instance) for obj in objectives),
                                 key=str))
            obounds = {}
            for objf in objfs:
                obounds[f"{objf}{_OBJECTIVE_LOWER}"] = objf.lower_bound()
                obounds[f"{objf}{_OBJECTIVE_UPPER}"] = objf.upper_bound()
            row = ({key: bin_bounds[key](instance)
                    for key in sorted(bin_bounds.keys())}, objfs,
                   immutable_mapping(obounds))
        if cache is not None:
            cache[instance.name] = row
        return PackingResult(
            end_result=end_result,
            n_items=instance.n_items,
            n_different_items=instance.n_different_items,
            bin_width=instance.bin_width, bin_height=instance.bin_height,
            objectives={str(objf): cast(Objective, objf).evaluate(packing)
                        for objf in row[1]},
            objective_bounds=row[2],
            bin_bounds=row[0])

    @staticmethod
    def from_single_log(  # pylint: disable=W0102
            file: str,
            objectives: Iterable[Callable[[Instance], Objective]] =
            DEFAULT_OBJECTIVES,
            bin_bounds: Mapping[str, Callable[[Instance], int]] =
            _DEFAULT_BOUNDS,
            cache: Mapping[str, tuple[Mapping[str, int], tuple[
                Objective, ...], Mapping[str, int | float]]] | None =
            None) -> "PackingResult":
        """
        Create a `PackingResult` from a file.

        :param file: the file path
        :param objectives: the objective function factories
        :param bin_bounds: the bounds computing functions
        :param cache: a cache that can store stuff if this function is to be
            called repeatedly
        :return: the packing result
        """
        file_path = Path.file(file)
        end_results: Final[list[EndResult]] = []
        EndResult.from_logs(file_path, end_results.append)
        if len(end_results) != 1:
            raise ValueError(
                f"needs one end result record in file {file_path!r}, "
                f"but got {end_results}.")

        packing = Packing.from_log(file_path)
        if not isinstance(packing, Packing):
            raise type_error(packing, f"packing from {file!r}", Packing)
        return PackingResult.from_packing_and_end_result(
            end_result=end_results[0], packing=packing,
            objectives=objectives, bin_bounds=bin_bounds, cache=cache)

    @staticmethod
    def from_logs(  # pylint: disable=W0102
            directory: str,
            collector: Callable[["PackingResult"], None],
            objectives: Iterable[Callable[[Instance], Objective]] =
            DEFAULT_OBJECTIVES,
            bin_bounds: Mapping[str, Callable[[Instance], int]]
            = _DEFAULT_BOUNDS) -> None:
        """
        Parse a directory recursively to get all packing results.

        :param directory: the directory to parse
        :param collector: the collector for receiving the results
        :param objectives: the objective function factories
        :param bin_bounds: the bin bounds calculators
        """
        _LogParser(collector, objectives, bin_bounds).parse_dir(directory)

    @staticmethod
    def to_csv(results: Iterable["PackingResult"], file: str) -> Path:
        """
        Write a sequence of packing results to a file in CSV format.

        :param results: the end results
        :param file: the path
        :return: the path of the file that was written
        """
        path: Final[Path] = Path.path(file)
        logger(f"Writing packing results to CSV file {path!r}.")
        Path.path(os.path.dirname(path)).ensure_dir_exists()

        # get a nicely sorted view on the results
        use_results = sorted(results)

        # get the names of the bounds and objectives
        bin_bounds_set: set[str] = set()
        objectives_set: set[str] = set()
        needs_encoding: bool = False
        needs_max_fes: bool = False
        needs_max_ms: bool = False
        needs_goal_f: bool = False
        for pr in use_results:
            bin_bounds_set.update(pr.bin_bounds.keys())
            objectives_set.update(pr.objectives.keys())
            er = pr.end_result
            if er.encoding is not None:
                needs_encoding = True
            if er.max_fes is not None:
                needs_max_fes = True
            if er.max_time_millis is not None:
                needs_max_ms = True
            if (er.goal_f is not None) and (isfinite(er.goal_f)):
                needs_goal_f = True
        bin_bounds: Final[list[str]] = sorted(bin_bounds_set)
        objectives: Final[list[str]] = sorted(objectives_set)

        line: list[str] = [_HEADER_1]
        if needs_encoding:
            line.append(KEY_ENCODING)
        line.append(_HEADER_2)
        line.extend(bin_bounds)
        for ob in objectives:
            line.append(f"{ob}{_OBJECTIVE_LOWER}{CSV_SEPARATOR}"
                        f"{ob}{_OBJECTIVE_UPPER}")
        line.append(_HEADER_3)
        if needs_goal_f:
            line.append(KEY_GOAL_F)
        if needs_max_fes:
            line.append(KEY_MAX_FES)
        if needs_max_ms:
            line.append(KEY_MAX_TIME_MILLIS)
        line.extend(objectives)

        with path.open_for_write() as out:
            out.write(CSV_SEPARATOR.join(line))
            line.clear()
            out.write("\n")
            for pr in use_results:
                e: EndResult = pr.end_result
                line.append(f"{e.algorithm}{CSV_SEPARATOR}"
                            f"{e.instance}{CSV_SEPARATOR}{e.objective}")
                if needs_encoding:
                    line.append("" if e.encoding is None else e.encoding)
                line.append(f"{pr.n_items}{CSV_SEPARATOR}"
                            f"{pr.n_different_items}{CSV_SEPARATOR}"
                            f"{pr.bin_width}{CSV_SEPARATOR}"
                            f"{pr.bin_height}")
                for bb in bin_bounds:
                    line.append(
                        str(pr.bin_bounds[bb]) if bb in pr.bin_bounds else "")
                for ob in objectives:
                    ox = f"{ob}{_OBJECTIVE_LOWER}"
                    line.append(num_to_str(pr.objective_bounds[ox])
                                if ox in pr.objective_bounds else "")
                    ox = f"{ob}{_OBJECTIVE_UPPER}"
                    line.append(num_to_str(pr.objective_bounds[ox])
                                if ox in pr.objective_bounds else "")
                line.append(
                    f"{hex(e.rand_seed)}{CSV_SEPARATOR}"
                    f"{num_to_str(e.best_f)}{CSV_SEPARATOR}"
                    f"{e.last_improvement_fe}{CSV_SEPARATOR}"
                    f"{e.last_improvement_time_millis}{CSV_SEPARATOR}"
                    f"{e.total_fes}{CSV_SEPARATOR}"
                    f"{e.total_time_millis}")
                if needs_goal_f:
                    line.append(intfloatnone_to_str(e.goal_f))
                if needs_max_fes:
                    line.append(intnone_to_str(e.max_fes))
                if needs_max_ms:
                    line.append(intnone_to_str(e.max_time_millis))
                for ob in objectives:
                    line.append(num_to_str(pr.objectives[ob])
                                if ob in pr.objectives else "")
                out.write(CSV_SEPARATOR.join(line))
                line.clear()
                out.write("\n")

        logger(f"Done writing end results to CSV file {path!r}.")
        return path

    @staticmethod
    def from_csv(file: str,
                 collector: Callable[["PackingResult"], None]) -> None:
        """
        Load the packing results from a CSV file.

        :param file: the file to read from
        :param collector: the collector for the results
        """
        path = Path.file(file)
        if not callable(collector):
            raise type_error(collector, "collector", call=True)

        with (path.open_for_read() as stream):
            header = stream.readline()
            if not isinstance(header, str):
                raise type_error(header, f"{file!r}[0]", str)

            idx_algorithm: int = -1
            idx_instance: int = -1
            idx_objective: int = -1
            idx_encoding: int = -1
            idx_seed: int = -1
            idx_li_fe: int = -1
            idx_li_ms: int = -1
            idx_best_f: int = -1
            idx_tt_fe: int = -1
            idx_tt_ms: int = -1
            idx_goal_f: int = -1
            idx_max_fes: int = -1
            idx_max_ms: int = -1
            idx_n_items: int = -1
            idx_n_different: int = -1
            idx_bin_width: int = -1
            idx_bin_height: int = -1
            idx_bounds: dict[str, int] = {}
            idx_objective_bounds: dict[str, int] = {}
            idx_objectives: dict[str, int] = {}

            for i, cellstr in enumerate(header.strip().split(CSV_SEPARATOR)):
                cell = cellstr.strip()
                if cell == KEY_N_ITEMS:
                    idx_n_items = i
                elif cell == KEY_N_DIFFERENT_ITEMS:
                    idx_n_different = i
                elif cell == KEY_BIN_WIDTH:
                    idx_bin_width = i
                elif cell == KEY_BIN_HEIGHT:
                    idx_bin_height = i
                elif cell == KEY_ALGORITHM:
                    idx_algorithm = i
                elif cell == KEY_INSTANCE:
                    idx_instance = i
                elif cell == KEY_OBJECTIVE_FUNCTION:
                    idx_objective = i
                elif cell == KEY_ENCODING:
                    idx_encoding = i
                elif cell == KEY_RAND_SEED:
                    idx_seed = i
                elif cell == KEY_LAST_IMPROVEMENT_FE:
                    idx_li_fe = i
                elif cell == KEY_LAST_IMPROVEMENT_TIME_MILLIS:
                    idx_li_ms = i
                elif cell == KEY_BEST_F:
                    idx_best_f = i
                elif cell == KEY_TOTAL_FES:
                    idx_tt_fe = i
                elif cell == KEY_TOTAL_TIME_MILLIS:
                    idx_tt_ms = i
                elif cell == KEY_GOAL_F:
                    idx_goal_f = i
                elif cell == KEY_MAX_FES:
                    idx_max_fes = i
                elif cell == KEY_MAX_TIME_MILLIS:
                    idx_max_ms = i
                elif cell.startswith(_BINS_START):
                    idx_bounds[cell] = i
                elif cell.endswith((_OBJECTIVE_LOWER, _OBJECTIVE_UPPER)):
                    idx_objective_bounds[cell] = i
                else:
                    idx_objectives[cell] = i

            if idx_n_items < 0:
                raise ValueError(
                    f"Missing key {KEY_N_ITEMS!r} in header "
                    f"{header!r} of file {file!r}.")
            if idx_n_different < 0:
                raise ValueError(
                    f"Missing key {KEY_N_DIFFERENT_ITEMS!r} in header "
                    f"{header!r} of file {file!r}.")
            if idx_bin_width < 0:
                raise ValueError(
                    f"Missing key {KEY_BIN_WIDTH!r} in header "
                    f"{header!r} of file {file!r}.")
            if idx_bin_height < 0:
                raise ValueError(
                    f"Missing key {KEY_BIN_HEIGHT!r} in header "
                    f"{header!r} of file {file!r}.")
            if idx_algorithm < 0:
                raise ValueError(
                    f"Missing key {KEY_ALGORITHM!r} in header "
                    f"{header!r} of file {file!r}.")
            if idx_instance < 0:
                raise ValueError(
                    f"Missing key {KEY_INSTANCE!r} in header "
                    f"{header!r} of file {file!r}.")
            if idx_objective < 0:
                raise ValueError(
                    f"Missing key {KEY_OBJECTIVE_FUNCTION!r} in header "
                    f"{header!r} of file {file!r}.")
            if idx_seed < 0:
                raise ValueError(
                    f"Missing key {KEY_RAND_SEED!r} in "
                    f"header {header!r} of file {file!r}.")
            if idx_li_fe < 0:
                raise ValueError(
                    f"Missing key {KEY_LAST_IMPROVEMENT_FE!r} in header "
                    f"{header!r} of file {file!r}.")
            if idx_li_ms < 0:
                raise ValueError(
                    f"Missing key {KEY_LAST_IMPROVEMENT_TIME_MILLIS!r} in "
                    f"header {header!r} of file {file!r}.")
            if idx_best_f < 0:
                raise ValueError(
                    f"Missing key {KEY_BEST_F!r} in header "
                    f"{header!r} of file {file!r}.")
            if idx_tt_fe < 0:
                raise ValueError(
                    f"Missing key {KEY_TOTAL_FES!r} in "
                    f"header {header!r} of file {file!r}.")
            if idx_tt_ms < 0:
                raise ValueError(
                    f"Missing key {KEY_TOTAL_TIME_MILLIS!r} in "
                    f"header {header!r} of file {file!r}.")

            objectives: list[str] = sorted(idx_objectives.keys())
            if len(objectives) <= 0:
                raise ValueError(f"no objectives found in file {file!r}.")
            for kk in objectives:
                if f"{kk}{_OBJECTIVE_LOWER}" not in idx_objective_bounds:
                    raise ValueError(
                        f"objective {kk!r} has no lower bound in header "
                        f"{header} of file {file!r}.")
                if f"{kk}{_OBJECTIVE_UPPER}" not in idx_objective_bounds:
                    raise ValueError(
                        f"objective {kk!r} has no upper bound in header "
                        f"{header} of file {file!r}.")
            if len(idx_objective_bounds) != (2 * len(objectives)):
                raise ValueError(
                    f"inconsistent bounds {idx_objective_bounds.keys()} in "
                    f"header {header} of file {file!r}.")

            bounds: list[str] = sorted(idx_bounds.keys())
            if len(bounds) <= 0:
                raise ValueError(f"no bounds found in file {file!r}.")

            for line in stream:
                splt = line.strip().split(CSV_SEPARATOR)
                encoding: str | None
                if idx_encoding < 0:
                    encoding = None
                else:
                    encoding = splt[idx_encoding].strip()
                    if len(encoding) <= 0:
                        encoding = None

                er = EndResult(
                    splt[idx_algorithm].strip(),  # algorithm
                    splt[idx_instance].strip(),  # instance
                    splt[idx_objective].strip(),  # objective
                    encoding,  # encoding
                    int((splt[idx_seed])[2:], 16),  # rand seed
                    str_to_intfloat(splt[idx_best_f]),  # best_f
                    int(splt[idx_li_fe]),  # last_improvement_fe
                    int(splt[idx_li_ms]),  # last_improvement_time_millis
                    int(splt[idx_tt_fe]),  # total_fes
                    int(splt[idx_tt_ms]),  # total_time_millis
                    None if idx_goal_f < 0 else
                    str_to_intfloatnone(splt[idx_goal_f]),  # goal_f
                    None if idx_max_fes < 0 else
                    str_to_intnone(splt[idx_max_fes]),  # max_fes
                    None if idx_max_ms < 0 else
                    str_to_intnone(splt[idx_max_ms]))  # max_time_millis

                the_objectives: dict[str, int | float] = {
                    on: str_to_intfloat(splt[idx_objectives[on]])
                    for on in objectives}
                the_bin_bounds: dict[str, int] = {
                    on: int(splt[idx_bounds[on]]) for on in bounds}
                the_objective_bounds: dict[str, int | float] = {
                    on: str_to_intfloat(splt[idxx])
                    for on, idxx in idx_objective_bounds.items()}

                collector(PackingResult(
                    n_items=int(splt[idx_n_items]),
                    n_different_items=int(splt[idx_n_different]),
                    bin_width=int(splt[idx_bin_width]),
                    bin_height=int(splt[idx_bin_height]),
                    end_result=er,
                    objectives=the_objectives,
                    objective_bounds=the_objective_bounds,
                    bin_bounds=the_bin_bounds))


class _LogParser(LogParser):
    """The internal log parser class."""

    def __init__(self, collector: Callable[["PackingResult"], None],
                 objectives: Iterable[Callable[[Instance], Objective]],
                 bin_bounds: Mapping[str, Callable[[Instance], int]]) -> None:
        """
        Parse a directory recursively to get all packing results.

        :param collector: the collector for receiving the results
        :param objectives: the objective function factories
        :param bin_bounds: the bin bounds calculators
        """
        super().__init__()
        if not callable(collector):
            raise type_error(collector, "collector", call=True)
        if not isinstance(objectives, Iterable):
            raise type_error(objectives, "objectives", Iterable)
        if not isinstance(bin_bounds, Mapping):
            raise type_error(bin_bounds, "bin_bounds", Mapping)
        #: the internal collector
        self.__collector: Final[Callable[[PackingResult], None]] = collector
        #: the objectives holder
        self.__objectives: Final[
            Iterable[Callable[[Instance], Objective]]] = objectives
        #: the bin bounds
        self.__bin_bounds: Final[
            Mapping[str, Callable[[Instance], int]]] = bin_bounds
        #: the internal cache
        self.__cache: Final[Mapping[
            str, tuple[Mapping[str, int], tuple[
                Objective, ...], Mapping[str, int | float]]]] = {}

    def parse_file(self, path: str) -> bool:
        """
        Parse a log file.

        :param path: the path to the log file
        :return: `True`
        """
        self.__collector(PackingResult.from_single_log(
            path, self.__objectives, self.__bin_bounds, self.__cache))
        return True


# Evaluate an experiment from the command line

# Run log files to end results if executed as script
if __name__ == "__main__":
    parser: Final[argparse.ArgumentParser] = argparser(
        __file__,
        "Convert log files for the bin packing experiment to a CSV file.",
        "Re-evaluate all results based on different objective functions.")
    parser.add_argument(
        "source", nargs="?", default="./results",
        help="the location of the experimental results, i.e., the root folder "
             "under which to search for log files", type=Path.path)
    parser.add_argument(
        "dest", help="the path to the end results CSV file to be created",
        type=Path.path, nargs="?", default="./evaluation/end_results.txt")
    args: Final[argparse.Namespace] = parser.parse_args()

    packing_results: Final[list[PackingResult]] = []
    PackingResult.from_logs(args.source, packing_results.append)
    PackingResult.to_csv(packing_results, args.dest)



"""A logger for results gathered from ODE integration to a text file."""

from contextlib import AbstractContextManager
from io import TextIOBase
from typing import Callable, Final

import numpy as np
from moptipy.utils.console import logger
from moptipy.utils.logger import CSV_SEPARATOR
from moptipy.utils.path import Path
from moptipy.utils.strings import float_to_str


class ResultsLog(AbstractContextManager):
    """
    A class for logging results via `multi_run_ode`.

    Function :func:`moptipyapps.dynamic_control.ode.multi_run_ode` can pass
    its results to various output generating procedures. This class here
    offers a procedure for writing them to a log file.

    >>> def projectile(position, ttime, ctrl, out):
    ...     out[0] = 70.71067811865474
    ...     out[1] = 70.71067811865474 - ttime * 9.80665
    >>> param: np.ndarray = np.array([1])   # ignore
    >>> def contrl(position, ttime, params, dest):
    ...     dest[0] = params[0]  #  controller that returns param
    >>> from io import StringIO
    >>> from moptipyapps.dynamic_control.ode import multi_run_ode
    >>> with StringIO() as sio:
    ...     with ResultsLog(2, sio) as log:
    ...         multi_run_ode([np.array([0.0, 1.0])],
    ...                       [np.array([1.0, 1.0])],
    ...                       log.collector, projectile, contrl, param,
    ...                       1, 10000, 14.890, 10000, 14.961)
    ...         x=sio.getvalue()
    >>> tt = x.split()
    >>> print(tt[0])
    figureOfMerit;totalTime;nSteps;start0;start1;end0;end1
    >>> for y in tt[1:]:
    ...     print(";".join(f"{float(v):.3f}" for v in y.split(";")))
    403374.924;14.890;10000.000;0.000;1.000;1052.882;-33.244
    407810.750;14.961;10000.000;1.000;1.000;1058.902;-38.616
    """

    def __init__(self, state_dim: int, out: TextIOBase | str) -> None:
        """
        Create the test results logger.

        :param state_dim: the state dimension
        :param out: the output destination
        """
        super().__init__()
        if isinstance(out, str):
            pp = Path.path(out)
            logger(f"logging data to file {pp!r}.")
            out = pp.open_for_write()
        #: the internal output destination
        self.__out: TextIOBase = out
        #: the state dimension
        self.__state_dim: Final[int] = state_dim

    def collector(self, index: int, ode: np.ndarray,
                  j: float, time: float) -> None:
        """
        Log the result of a multi-ode run.

        :param index: the index of the result
        :param ode: the ode result matrix
        :param j: the figure of merit
        :param time: the time value
        """
        if self.__out is None:
            raise ValueError("Already closed output destination!")
        out_str: Final[Callable[[str], int]] = self.__out.write

        state_dim: Final[int] = self.__state_dim
        if index <= 0:
            out_str(
                f"figureOfMerit{CSV_SEPARATOR}totalTime{CSV_SEPARATOR}nSteps")
            for i in range(state_dim):
                out_str(f"{CSV_SEPARATOR}start{i}")
            for i in range(state_dim):
                out_str(f"{CSV_SEPARATOR}end{i}")
            out_str("\n")

        out_str(float_to_str(float(j)))
        out_str(CSV_SEPARATOR)
        out_str(float_to_str(float(time)))
        out_str(CSV_SEPARATOR)
        out_str(str(len(ode)))
        start = ode[0]
        for i in range(state_dim):
            out_str(f"{CSV_SEPARATOR}{float_to_str(float(start[i]))}")
        end = ode[-1]
        for i in range(state_dim):
            out_str(f"{CSV_SEPARATOR}{float_to_str(float(end[i]))}")
        out_str("\n")

    def __exit__(self, _, __, ___) -> None:
        """
        Close this context manager.

        :param _: the exception type; ignored
        :param __: the exception value; ignored
        :param ___: the exception whatever; ignored
        """
        if self.__out is not None:
            try:
                self.__out.close()
            finally:
                self.__out = None



"""
A primitive integrator for systems of ordinary differential equations.

Many dynamic systems can be modeled as systems of ordinary differential
equations that govern their progress over time. Trying to find out in
which state such systems are at a given point in time means to integrate
these equations until that point in time (starting from a starting state).

What we want to play around with, however, is synthesizing controllers.
In this case, the differential equations also merge the output of the
controller with the current state. If the controller behaves inappropriately,
this may make the system diverge, i.e., some of its state variables go to
infinity over time or sometimes rather quickly.

Using ODE integrators that compute the system state at pre-defined time steps
is thus cumbersome, as the system may have already exploded at these goal
times. Therefore, we perform ODE integration in several steps. First, we try
it the "normal" way. However, as soon as the system escapes the sane parameter
range, we stop. We then use the last point where the system was stable and the
first point where it escaped the reasonable range to estimate a new reasonable
end time for the integration. We do this until we finally succeed.

Thus, we can simulate a well-behaved system over a long time and an
ill-behaved system for a shorter time period. Neither system will diverge.

The following functions are provided:

- :func:`run_ode` executes a set of differential system equations and
  controller equations and returns an array with the system state, controller
  output, and time at the different interpolation steps.
- :func:`t_from_ode` returns the total time over which the result of
  :func:`run_ode` was simulated.
- :func:`j_from_ode` returns a figure of merit, i.e., a weighted sum of (a
  part of) the system state and the controller output from the result of
  :func:`run_ode`.
- :func:`diff_from_ode` extracts state difference information from the result
  of :func:`run_ode`.
"""

from math import fsum, inf
from typing import Callable, Final, Iterable, TypeVar

import numba  # type: ignore
import numpy as np
from scipy.integrate import RK45, DenseOutput  # type: ignore

#: the type variable for ODE controller parameters
T = TypeVar("T")


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def _is_ok(x: np.ndarray) -> bool:
    """
    Check whether all values in a vector are acceptable.

    A vector is "ok" if all of its elements are from the finite range
    `(-1e10, 1e10)`.  Anything else indicates that we are somehow moving
    out of the reasonable bounds.

    :param x: the vector
    :return: `True` if all values are OK, `False` otherwise
    """
    for xx in x:  # noqa: SIM110
        if not -1e10 < xx < 1e10:
            return False
    return True


class __IntegrationState:
    """
    The internal integrator state class.

    This class serves two purposes. First, it encapsulates the system
    equations and the controller equations such that they can be called as a
    unit. Second, it raises an alert if the system escapes the "OK" state,
    i.e., diverge towards infinity. As long as everything is OK,
    :attr:`is_ok` will remain `True`. But if either the controller or the
    system state escapes the acceptable interval, it becomes `False`.
    In that case, :attr:`max_ok_t` holds the highest `t` value at which
    the system and controller were in an acceptable state and
    :attr:`min_error_t` holds the smallest `t` value at which that was not
    the case. We can assume that somewhere inbetween lies the last moment we
    can find the system in a sane state.
    """

    def __init__(self,
                 equations: Callable[[
                     np.ndarray, float, np.ndarray, np.ndarray], None],
                 controller: Callable[[
                     np.ndarray, float, T, np.ndarray], None],
                 parameters: T, controller_dim: int) -> None:
        """
        Create the integrator.

        :param equations: the differential system
        :param controller: the controller function
        :param parameters: the controller parameters
        :param controller_dim: the dimension of the controller result
        """
        self.__equations: Final[Callable[[
            np.ndarray, float, np.ndarray, np.ndarray], None]] = equations
        self.__controller: Final[Callable[[
            np.ndarray, float, T, np.ndarray], None]] = controller
        self.__parameters: Final[T] = parameters
        self.__ctrl: Final[np.ndarray] = np.empty(controller_dim)
        self.max_ok_t: float = -inf
        self.min_error_t: float = inf
        self.is_ok: bool = True

    def init(self) -> None:
        """Prepare the system for integration."""
        self.max_ok_t = -inf
        self.min_error_t = inf
        self.is_ok = True

    def f(self, t: float, state: np.ndarray) -> np.ndarray:
        """
        Compute the differential at the given state and time.

        First, we invoke the controller function at the state and time. Then
        we pass the controller vector to the differential equations to update
        the system state.
        This function also checks if the state or control go out of bounds
        and if they do, it sets the :attr:`is_ok` to `False`.

        :param t: the time
        :param state: the state
        :return: the differential
        """
        out: Final[np.ndarray] = np.zeros_like(state)  # allocate output vec
        ctrl: Final[np.ndarray] = self.__ctrl  # the controller vector

        # invoke the controller
        self.__controller(state, t, self.__parameters, ctrl)
        ok: bool = _is_ok(ctrl)  # is the controller vector ok?
        if ok:  # if yes, let's invoke the state update equations
            self.__equations(state, t, ctrl, out)
            ok = _is_ok(out)  # and check if their result is ok
        if ok:  # is it ok?
            if self.max_ok_t < t < self.min_error_t:
                self.max_ok_t = t
        else:  # no: there was some error, either in state or controller
            self.is_ok = False  # then we are no longer OK
            if t < self.min_error_t:  # update the minimum error time
                self.min_error_t = t
            if self.max_ok_t > t:  # what? an earlier error?
                self.max_ok_t = np.nextafter(t, -inf)  # ok, reset ok time

        return out


def run_ode(starting_state: np.ndarray,
            equations: Callable[[
                np.ndarray, float, np.ndarray, np.ndarray], None],
            controller: Callable[[np.ndarray, float, T, np.ndarray], None],
            parameters: T, controller_dim: int = 1,
            steps: int = 5000, max_time: float = 50.0) -> np.ndarray:
    """
    Simulate a set of controlled differential system.

    The system begins in the starting state stored in the vector
    `starting_state`. In each time step, first, the `controller` is invoked.
    It receives the current system state vector as input, the current time
    `t`, its parameters (`parameters`), and an output array to store its
    computed control values into. This array has dimension `controller_dim`,
    which usually is `1`. Then, the function `system` will be called and
    receives the current state vector, the time step `t`, and the controller
    output as input, as well as an output array to store the result of the
    differential into. This output array has the same dimension as the state
    vector.

    Now the `run_ode` function simulates such a system over `steps` time steps
    over the closed interval `[0, max_time]`. If both the system and the
    controller are well-behaved, then the output array will contain `steps`
    rows with the state, controller, and time information of each step. If
    the system diverges at some point in time but we can simulate it
    reasonably well before that, then we try to simulate `steps`, but on a
    shorter time frame. If even that fails, you will get a single row output
    with `1e100` as the controller value.

    This function returns a matrix where each row corresponds to a simulated
    time step. Each row contains three components in a concatenated fashion:
    1. the state vector,
    2. the control vector,
    3. the time value

    :param starting_state: the starting
    :param equations: the differential system
    :param controller: the controller function
    :param parameters: the controller parameters
    :param controller_dim: the dimension of the controller result
    :param steps: the number of steps to simulate
    :param max_time: the maximum time to simulate for
    :returns: a matrix where each row represents a point in time, composed of
        the current state, the controller output, and the length of the time
        slice

    If we simulate the flight of a projectile with our ODE execution, then
    both the flight time as well as the flight length are about 0.12% off from
    what the mathematical solution of the flight system prescribe. That's
    actually not bad for a crude and fast integration method...

    >>> v = 100.0
    >>> angle = np.deg2rad(45.0)
    >>> v_x = v * np.cos(angle)
    >>> print(f"{v_x:.10f}")
    70.7106781187
    >>> v_y = v * np.sin(angle)
    >>> print(f"{v_y:.10f}")
    70.7106781187
    >>> def projectile(position, ttime, ctrl, out):
    ...     out[0] = 70.71067811865474
    ...     out[1] = 70.71067811865474 - ttime * 9.80665
    >>> param = 0.0   # ignore
    >>> def contrl(position, ttime, params, dest):
    ...     dest[0] = 0.0  #  controller that does nothing
    >>> strt = np.array([0.0, 1.0])
    >>> ode = run_ode(strt, projectile, contrl, param, 1, 10000)
    >>> print(len(ode))
    10000
    >>> time_of_flight = 2 * v_y / 9.80665
    >>> print(f"{time_of_flight:.10f}")
    14.4209649817
    >>> travel_distance_x = time_of_flight * v_x
    >>> print(f"{travel_distance_x:.10f}")
    1019.7162129779
    >>> idx = np.argwhere(ode[:, 1] <= 0.0)[0][0]
    >>> print(idx)
    2887
    >>> print(f"{ode[idx - 1, 0]:.10f}")
    1020.4571309653
    >>> print(f"{ode[idx, 0]:.10f}")
    1020.8107197148
    >>> print(f"{ode[idx - 1, -1]:.10f}")
    14.4314431443
    >>> print(f"{ode[idx, -1]:.10f}")
    14.4364436444
    >>> print(ode[-1, -1])
    50.0

    >>> def contrl2(position, ttime, params, dest):
    ...     dest[0] = 1e50  #  controller that is ill-behaved
    >>> run_ode(strt, projectile, contrl2, param, 1, 10000)
    array([[0.e+000, 1.e+000, 1.e+100, 0.e+000]])

    >>> def contrl3(position, ttime, params, dest):
    ...     dest[0] = 1e50 if ttime > 10 else 0.0  # diverging controller
    >>> ode = run_ode(strt, projectile, contrl3, param, 1, 10000)
    >>> print(len(ode))
    10000
    >>> print(ode[-1])
    [690.10677249 224.06765771   0.           9.75958357]

    >>> def projectile2(position, ttime, ctrl, out):
    ...     out[:] = 0
    >>> ode = run_ode(strt, projectile2, contrl, param, 1, 10000)
    >>> print(len(ode))
    10000
    >>> print(ode[-1])
    [ 0.  1.  0. 50.]
    >>> ode = run_ode(strt, projectile2, contrl3, param, 1, 10000)
    >>> print(len(ode))
    10000
    >>> print(ode[-1])
    [0.         1.         0.         9.41234557]
    """
    func_state: Final[__IntegrationState] = __IntegrationState(
        equations, controller, parameters, controller_dim)
    func_call: Final[Callable[[float, np.ndarray], np.ndarray]] = func_state.f
    denses: Final[list[DenseOutput]] = []
    n: Final[int] = len(starting_state)
    dim: Final[int] = n + controller_dim + 1

    cycle: int = 0
    while True:  # loop until we have a sane integration over a sane range
        cycle += 1
        # first, we reset all the state information
        func_state.init()  # reset the function state
        denses.clear()  # always discard all interpolators, if there are any

        # then we create the integrator for the time range that we simulate
        integration = RK45(
            fun=func_call, t0=0.0, y0=starting_state, t_bound=max_time,
            max_step=steps)
        is_finished: bool = False

        # perform the integration and collect all the points at which stuff
        # was computed
        while True:  # iteratively add interpolation points
            integration.step()  # do the integration step
            if not func_state.is_ok:
                break  # some out-of-bounds thing happened! quit!
            is_finished = integration.status == "finished"
            is_running: bool = integration.status == "running"
            if is_finished or is_running:  # keep taking interpolator
                denses.append(integration.dense_output())
                if is_finished:
                    break  # we are finished, so we quit and build output
                continue  # more integration to do, so we go on
            break  # if we get here, there was an error: quit

        if is_finished and func_state.is_ok:
            # if we get here, everything looks fine so far, so we can try
            # to build the output
            result: np.ndarray = np.zeros((steps, dim))
            point: np.ndarray = result[0]
            point[0:n] = starting_state

            # we now compute all the points by using the interpolation
            # we start with the first point
            controller(starting_state, 0.0, parameters, point[n:-1])
            if not _is_ok(point):
                break
            result[:, -1] = np.linspace(0.0, max_time, steps)

            j: int = 0  # the index of the dense interpolator to use
            n_dense: int = len(denses)  # the number of interpolators
            t: float = 0.0  # the current time
            dense: DenseOutput = denses[j]  # the current interpolator
            for point in result[1:]:  # for each of the remaining points...
                last_time: float = t  # remember the last successful point
                t = point[-1]  # get the time value

                # we now need to find the right interpolator if we have left
                # the range of the current interpolator
                while not (dense.t_min <= t <= dense.t_max):
                    j += 1  # step counter
                    if j >= n_dense:   # we have left the interpolation range??
                        func_state.max_ok_t = last_time  # so we need to adjust
                        func_state.min_error_t = t
                        is_finished = False  # and try the whole thing again
                        break
                    dense = denses[j]  # pick next interpolator

                if is_finished:  # if we get here, we got a right interpolator
                    point[0:n] = dense(t)  # so we can interpolate the state
                    controller(point[0:n], t, parameters, point[n:-1])
                    if not _is_ok(point):  # is there an error in the vector?
                        func_state.max_ok_t = last_time  # last ok time
                        func_state.min_error_t = t  # error time
                        is_finished = False  # we need to quit and try again
                        break  # stop inner loop

            if is_finished:  # did we succeed?
                return result  # yes! return the result.

        # if we arrive here, things went wrong somehow.
        # this means that we should reduce the maximum runtime
        if (cycle < 3) and (func_state.max_ok_t < func_state.min_error_t):
            max_time = np.nextafter(min(func_state.min_error_t, (
                0.8 * func_state.max_ok_t) + (0.2 * func_state.min_error_t)),
                -inf)
        else:  # the small reductions did not work out well ... reduce rapidly
            max_time = np.nextafter(0.7 * min(func_state.max_ok_t,
                                              max_time), -inf)

        if (cycle > 4) or (max_time <= 1e-10):
            break  # if we get here, everything seems so pointless...

    # the default error result
    result = np.zeros((1, dim))
    result[0, 0:n] = starting_state
    result[0, n:-1] = 1e100
    result[0, -1] = 0.0
    return result


@numba.njit(cache=True, inline="always", fastmath=False, boundscheck=False)
def __j_from_ode_compute(ode: np.ndarray, state_dim: int,
                         use_state_dims: int,
                         gamma: float,
                         dest: np.ndarray) -> None:
    """
    Prepare the input array for the figure of merit computation.

    The `ode` matrix contains one row for each time step.
    The row is composed of the current state, the current controller output,
    and the current time.
    However, all systems of a given simulation start in the same initial
    state, so it makes little sense to include this initial state into the
    figure of merit computation. It does make sense to include the control
    output at that moment, though, because it contributes to the next state.
    It also makes no sense to count in the last row of the ODE computation,
    because this is the final system state and the system will not spend any
    time in it in the simulation.

    :param ode: the output array from the ODE simulation
    :param state_dim: the state dimension
    :param use_state_dims: the dimension until which the state is used
    :param gamma: the weight for the control variable
    :param dest: the destination array

    >>> od = np.array([[1, 2, 3, 4, 0],
    ...                [5, 6, 7, 8, 1],
    ...                [9, 6, 4, 3, 3],
    ...                [7, 4, 2, 1, 7]])
    >>> sta_dim = 3
    >>> dst = np.empty((od.shape[0] - 1) * (od.shape[1] - 1) - sta_dim)
    >>> __j_from_ode_compute(od, sta_dim, sta_dim, 0.1, dst)
    >>> print(dst)
    [  1.6  12.8  98.   72.   50.    3.6  64.  144.  324. ]
    >>> rs = np.array([0.1 * 1 * 4 * 4,
    ...                0.1 * 2 * 8 * 8, 2 * 7 * 7, 2 * 6 * 6, 2 * 5 * 5,
    ...                0.1 * 4 * 3 * 3, 4 * 4 * 4, 4 * 6 * 6, 4 * 9 * 9])
    >>> print(rs)
    [  1.6  12.8  98.   72.   50.    3.6  64.  144.  324. ]
    >>> u_sta_dim = 2
    >>> dst = np.empty((od.shape[0] - 1) * (
    ...     od.shape[1] - 1 - sta_dim + u_sta_dim) - u_sta_dim)
    >>> __j_from_ode_compute(od, sta_dim, u_sta_dim, 1.0, dst)
    >>> print(dst)
    [ 16. 128.  72.  50.  36. 144. 324.]
    """
    index: int = 0
    start: Final[int] = ode.shape[1] - 2

    # from now on, we compute the impact of the state and the controller
    add_state: bool = False
    last_row: np.ndarray = ode[0]
    for i in range(1, len(ode)):
        next_row: np.ndarray = ode[i]
        weight: float = next_row[-1] - last_row[-1]
        inner = start
        weight_01: float = weight * gamma
        while inner >= state_dim:
            v = last_row[inner]
            inner -= 1
            dest[index] = (v * v) * weight_01 if -1e100 < v < 1e100 else 1e100
            index += 1

        if add_state:
            inner = use_state_dims  # jump to the used state
            while inner > 0:
                inner -= 1
                v = last_row[inner]
                dest[index] = (v * v) * weight if -1e100 < v < 1e100 else 1e100
                index += 1

        add_state = True
        last_row = next_row


def t_from_ode(ode: np.ndarray) -> float:
    """
    Get the time sum from an ODE solution.

    The total time that we simulate a system depends on the behavior of the
    system.

    :param ode: the ODE solution, as return from :func:`run_ode`.
    :return: the total consumed time

    >>> od = np.array([[1, 2, 3, 4, 0.1],
    ...                [5, 6, 7, 8, 0.2],
    ...                [9, 6, 4, 3, 0.3],
    ...                [7, 4, 2, 1, 0.4]])
    >>> print(t_from_ode(od))
    0.4
    """
    return ode[-1, -1]


def j_from_ode(ode: np.ndarray, state_dim: int,
               use_state_dims: int = -1,
               gamma: float = 0.1) -> float:
    """
    Compute the original figure of merit from an ODE array.

    The figure of merit is the sum of state variable squares plus 0.1 times
    the control variable squares. We disregard the state variable values of
    the starting states (because they are the same for all controllers on
    a given training case and because the control cannot influence them) and
    we also disregard the final state and final controller output (as there is
    no time slice associated with them, i.e., we only "arrive" in them but
    basically spent 0 time in them in our simulation).

    :param ode: the array returned by the ODE function, i.e., :func:`run_ode`
    :param state_dim: the state dimension
    :param use_state_dims: the dimension until which the state is used,
        `-1` for using the complete state
    :param gamma: the weight of the controller input
    :return: the figure of merit

    >>> od = np.array([[1, 2, 3, 4, 0],
    ...                [5, 6, 7, 8, 1],
    ...                [9, 6, 4, 3, 3],
    ...                [7, 4, 2, 1, 7]])
    >>> sta_dim = 3
    >>> print(f"{j_from_ode(od, sta_dim):.10f}")
    110.0000000000
    >>> print((1.6 + 12.8 + 98 + 72 + 50 + 3.6 + 64 + 144 + 324) / 7)
    110.0
    >>> sta_dim = 3
    >>> print(f"{j_from_ode(od, 3, 2, 0.5):.10f}")
    97.1428571429
    >>> print((8 + 64 + 72 + 50 + 18 + 144 + 324) / 7)
    97.14285714285714
    """
    if len(ode) <= 1:
        return 1e200
    # The used state dimension could be equal to the state dimension or less.
    # If it is <= 0, then we use the complete state vector
    if use_state_dims <= 0:
        use_state_dims = state_dim
    dest: Final[np.ndarray] = np.empty((ode.shape[0] - 1) * (
        ode.shape[1] - 1 - state_dim + use_state_dims) - use_state_dims)
    __j_from_ode_compute(ode, state_dim, use_state_dims, gamma, dest)
    return fsum(dest) / ode[-1, -1]


def multi_run_ode(
        test_starting_states: Iterable[np.ndarray],
        training_starting_states: Iterable[np.ndarray],
        collector: Callable[[int, np.ndarray, float, float], None] | Iterable[
            Callable[[int, np.ndarray, float, float], None]],
        equations: Callable[[
            np.ndarray, float, np.ndarray, np.ndarray], None],
        controller: Callable[[np.ndarray, float, T, np.ndarray], None],
        parameters: T, controller_dim: int = 1,
        test_steps: int = 5000,
        test_time: float = 50.0,
        training_steps: int = 5000,
        training_time: float = 50.0,
        use_state_dims: int = -1, gamma: float = 0.1) -> None:
    """
    Invoke :func:`run_ode` multiple times and pass the result to `collector`.

    This function allows us to perform multiple runs of the differential
    equation simulator, using different starting points. It also allows us to
    distinguish training and test points and to assign them different numbers
    of steps. For each of them, :func:`run_ode` will be applied and the
    returned matrix is passed to the `collector` function.

    :param test_starting_states: the iterable of test starting states
    :param training_starting_states: the iterable of training starting states
    :param collector: the destination to receive the results, in the form of
        index, ode array, j, and t.
    :param equations: the differential system
    :param controller: the controller function
    :param parameters: the controller parameters
    :param controller_dim: the dimension of the controller result
    :param test_steps: the number of test steps to simulate
    :param test_time: the time limit for tests
    :param training_steps: the number of training steps to simulate
    :param training_time: the time limit for training
    :param use_state_dims: the dimension until which the state is used,
        `-1` for using the complete state
    :param gamma: the weight of the controller input
    """
    if not isinstance(collector, Iterable):
        collector = (collector, )
    index: int = 0
    for sp in test_starting_states:
        ode = run_ode(sp, equations, controller, parameters, controller_dim,
                      test_steps, test_time)
        for c in collector:
            c(index, ode, j_from_ode(ode, len(sp), use_state_dims, gamma),
              t_from_ode(ode))
        index += 1
    for sp in training_starting_states:
        ode = run_ode(sp, equations, controller, parameters, controller_dim,
                      training_steps, training_time)
        for c in collector:
            c(index, ode, j_from_ode(ode, len(sp), use_state_dims, gamma),
              t_from_ode(ode))
        index += 1


def diff_from_ode(ode: np.ndarray, state_dim: int) \
        -> tuple[np.ndarray, np.ndarray]:
    """
    Compute all the state+control vectors and the resulting differentials.

    This function returns two matrices. Each row of both matrices corresponds
    to a time slot. Each row in the first matrix holds the state vector and
    the control vector (that was computed by the controller). The
    corresponding row in the second matrix then holds the state differential
    resulting from the control vector being applied in the differential
    equations that govern the system state change.

    The idea is that this function basically provides the data that we would
    like to learn when training a surrogate model for a system: From the
    current state and the computed control vector, we want that our model can
    give us the resulting system differential. If we have such a model and it
    works reasonably well, then we could essentially plug this model into
    :func:`run_ode` instead of the original `equations` parameter.

    What this function does to compute the differential is to basically
    "invert" the dynamic weighting done by :func:`run_ode`. :func:`run_ode`
    starts in a given starting state `s`. It then computes the control vector
    `c` as a function of `s`, i.e., `c(s)`. Then, the equations of the dynamic
    system (see module :mod:`~moptipyapps.dynamic_control.system`) to compute
    the state differential `D=ds/dt` as a function of `c(s)` and `s`, i.e., as
    something like `D(s, c(s))`. The next step would be to update the state,
    i.e., to set `s=s+D(s, c(s))`. Unfortunately, this can make `s` go to
    infinity. So :func:`run_ode` will compute a dynamic weight `w` and do
    `s=s+w*D(s, c(s))`, where `w` is chosen such that the state vector `s`
    does not grow unboundedly. While `s` and `c(s)` and `w` are stored in one
    row of the result matrix of :func:`run_ode`, `s+w*D(s,c(s))` is stored as
    state `s` in the next row. So what this function here basically does is to
    subtract the old state from the next state and divide the result by `w` to
    get `D(s, c(s))`. `s` and `c(s)` are already available directly in the ODE
    result and `w` is not needed anymore.

    We then get the rows `s, c(s)` and `D(s, c(s))` in the first and second
    result matrix, respectively. This can then be used to train a system model
    as proposed in model :mod:`~moptipyapps.dynamic_control.system_model`.

    :param ode: the result of :func:`run_ode`
    :param state_dim: the state dimensions
    :returns: a tuple of the state+control vectors and the resulting
        state differential vectors

    >>> od = np.array([
    ...     [0, 0, 0, 0, 0, 0],  # state 0,0,0; control 0,0; time 0
    ...     [1, 2, 3, 4, 5, 1],  # state 1,2,3; control 4,5; time 1
    ...     [2, 3, 4, 5, 6, 3],  # state 2,3,4; control 5,6; time 3
    ...     [4, 6, 8, 7, 7, 7]])    # state 4,6,8; control 7,7, time 7
    >>> res = diff_from_ode(od, 3)
    >>> res[0]  # state and control vectors, time col and last row removed
    array([[0, 0, 0, 0, 0],
           [1, 2, 3, 4, 5],
           [2, 3, 4, 5, 6]])
    >>> res[1]  # (state[i + 1] - state[i]) / (time[i + 1] / time[i])
    array([[1.  , 2.  , 3.  ],
           [0.5 , 0.5 , 0.5 ],
           [0.5 , 0.75, 1.  ]])
    """
    return (ode[0:-1, 0:-1],
            np.diff(ode[:, 0:state_dim], 1, 0)
            / np.diff(ode[:, -1])[:, None])



"""
Poor man's Artificial Neural Networks.

Here, artificial neural networks (ANNs) are defined as plain mathematical
functions which are parameterized by their weights. The weights are subject
to black-box optimization and all together put into a single vector.
In other words, we do not use proper back-propagation learning or any other
sophisticated neural network specific training strategy. Instead, we treat the
neural networks as black boxes that can be parameterized using the weight
vector. Different ANN architectures have different weight vectors.
As activation functions, we use `arctan`.

The neural networks here are automatically generated via code generation
provided by module :mod:`~moptipyapps.dynamic_control.controllers.codegen`.
This allows us to have networks of arbitrary shape for arbitrary input and
output dimensions. Since code generation followed by on-the-fly compilation
via numba is time and memory consuming, we cache all neural network-based
instances of :class:`~moptipyapps.dynamic_control.controller.Controller`.
"""

from typing import Callable, Final, Iterable, cast

import numpy as np
from moptipy.utils.types import check_int_range, type_error

from moptipyapps.dynamic_control.controller import Controller
from moptipyapps.dynamic_control.controllers.codegen import CodeGenerator
from moptipyapps.dynamic_control.system import System


def make_ann(state_dims: int, control_dims: int, layers: list[int]) \
        -> Controller:
    """
    Dynamically create an ANN.

    :param state_dims: the state or input dimension
    :param control_dims: the output dimension
    :param layers: the sizes of the hidden layers
    :returns: the controller
    """
    state_dims = check_int_range(state_dims, "state_dims", 1, 100)
    control_dims = check_int_range(control_dims, "state_dims", 1, 100)
    if not isinstance(layers, list):
        raise type_error(layers, "layers", list)
    for layer in layers:
        check_int_range(layer, "layer", 1, 64)

    # we also try to cache the generated controllers
    description = "_".join(map(str, ([state_dims, control_dims, *layers])))
    description = f"__cache_{description}"
    if hasattr(make_ann, description):
        return cast(Controller, getattr(make_ann, description))

    code: Final[CodeGenerator] = CodeGenerator(
        "state: np.ndarray, _: float, params: np.ndarray, out: np.ndarray")
    params: int = 0  # the number of parameters
    var_count: int = 0  # the number of variables
    vars_in: list[str] = []  # the variables forming the current layer input
    vars_out: list[str] = []  # the variables forming the current layer output
    vars_cached: list[str] = []  # the variables cached for re-use
    write: Final[Callable[[str], None]] = code.write  # fast call
    writeln: Final[Callable[[str], None]] = code.writeln  # fast call

    # first, we cache the state vector into local variables
    for i in range(state_dims):
        vv = f"s{i}"
        vars_in.append(vv)
        writeln(f"{vv} = state[{i}]")

    # now we build the hidden layers of the network
    for layer in layers:
        for _ in range(layer):
            # allocate a variable for storing the current neuron's output
            if len(vars_cached) > 0:
                var = vars_cached.pop(-1)
            else:
                var_count += 1
                var = f"v{var_count}"
            vars_out.append(var)  # remember the variable
            write(f"{var} = np.arctan(params[{params}]")  # the bias
            params += 1
            for vv in vars_in:
                write(f" + params[{params}] * {vv}")  # input * weight
                params += 1
            writeln(")")
        vars_cached.extend(vars_in)  # old inputs ready for reuse
        vars_in.clear()  # inputs are no longer used
        vars_in, vars_out = vars_out, vars_in  # outputs become inputs

    # now we construct the output layer
    for i in range(control_dims):
        write(f"out[{i}] = params[{params}] * ")  # the multiplier
        params += 1
        write(f"np.arctan(params[{params}]")  # the bias
        params += 1
        for vv in vars_in:
            write(f" + params[{params}] * {vv}")  # input * weight
            params += 1
        writeln(")")

    result: Final[Controller] = Controller(
        f"ann_{'_'.join(map(str, layers))}" if len(layers) > 0 else "ann",
        state_dims, control_dims, params, code.build())
    # perform one test invocation
    result.controller(np.zeros(state_dims), 0.0, np.zeros(params),
                      np.empty(control_dims))
    setattr(make_ann, description, result)  # cache the controller
    return result


def anns(system: System) -> Iterable[Controller]:
    """
    Create poor man's ANNs fitting to a given system.

    Based on the dimensionality of the state space, we generate a set of ANNs
    with different numbers of layers and neurons. The weights of the neurons
    can then directly be optimized by a numerical optimization algorithm.
    This is, of course, probably much less efficient that doing some proper
    learning like back-propagation. However, it allows us to easily plug the
    ANNs into the same optimization routines as other controllers.

    :param system: the equations object
    :return: the ANNs
    """
    state_dims: Final[int] = system.state_dims
    control_dims: Final[int] = system.control_dims
    return (make_ann(state_dims, control_dims, []),
            make_ann(state_dims, control_dims, [1]),
            make_ann(state_dims, control_dims, [2]),
            make_ann(state_dims, control_dims, [3]),
            make_ann(state_dims, control_dims, [2, 2]),
            make_ann(state_dims, control_dims, [3, 2]))



"""A two-dimensional packing."""

from typing import Final

import numpy as np
from moptipy.api.component import Component
from moptipy.api.logging import SECTION_RESULT_Y, SECTION_SETUP
from moptipy.evaluation.log_parser import LogParser
from moptipy.utils.types import type_error

from moptipyapps.binpacking2d.instance import Instance

#: the index of the ID in a :class:`Packing` row
IDX_ID: Final[int] = 0
#: the index of the bin in a :class:`Packing` row
IDX_BIN: Final[int] = 1
#: the index of the left x coordinate in a :class:`Packing` row
IDX_LEFT_X: Final[int] = 2
#: the index of the bottom y coordinate in a :class:`Packing` row
IDX_BOTTOM_Y: Final[int] = 3
#: the index of the right x coordinate in a :class:`Packing` row
IDX_RIGHT_X: Final[int] = 4
#: the index of the top y coordinate in a :class:`Packing` row
IDX_TOP_Y: Final[int] = 5


class Packing(Component, np.ndarray):
    """
    A packing, i.e., a solution to an 2D bin packing instance.

    A packing is a two-dimensional numpy array. In each row, the position of
    one item is stored: 1. the item ID (starts at 1), 2. the bin into which
    the item is packaged (starts at 1), 3. the left x coordinate, 4. the
    bottom y coordinate, 5. the right x coordinate, 6. the top y coordinate.
    """

    #: the 2d bin packing instance
    instance: Instance
    #: the number of bins
    n_bins: int

    def __new__(cls, instance: Instance) -> "Packing":
        """
        Create a solution record for the 2D bin packing problem.

        :param cls: the class
        :param instance: the solution record
        """
        if not isinstance(instance, Instance):
            raise type_error(instance, "instance", Instance)
        obj: Final[Packing] = super().__new__(
            cls, (instance.n_items, 6), instance.dtype)
        #: the 2d bin packing instance
        obj.instance = instance
        #: the number of bins
        obj.n_bins = -1
        return obj

    def __str__(self):
        """
        Convert the packing to a compact string.

        :return: the compact string
        """
        return "\n".join(";".join(str(self[i, j]) for j in range(6))
                         for i in range(self.shape[0]))

    @staticmethod
    def from_log(file: str, instance: Instance | None = None) -> "Packing":
        """
        Load a packing from a log file.

        :param file: the log file path
        :param instance: the optional Packing instance: if `None` is provided,
            we try to load it from the resources
        :returns: the Packing
        """
        parser: Final[_PackingParser] = _PackingParser(instance)
        parser.parse_file(file)
        # noinspection PyProtectedMember
        res = parser._result
        if not isinstance(res, Packing):
            raise type_error(res, f"packing from {file!r}", Packing)
        return res


class _PackingParser(LogParser):
    """The log parser for loading packings."""

    def __init__(self, instance: Instance | None = None):
        """
        Create the packing parser.

        :param instance: the optional packing instance: if `None` is provided,
            we try to load it from the resources
        """
        super().__init__()
        if (instance is not None) and (not isinstance(instance, Instance)):
            raise type_error(instance, "instance", Instance)
        #: the internal instance
        self.__instance: Instance | None = instance
        #: the internal section mode: 0=none, 1=setup, 2=y
        self.__sec_mode: int = 0
        #: the packing string
        self.__packing_str: str | None = None
        #: the result packing
        self._result: Packing | None = None

    def start_section(self, title: str) -> bool:
        """Start a section."""
        super().start_section(title)
        self.__sec_mode = 0
        if title == SECTION_SETUP:
            if self.__instance is None:
                self.__sec_mode = 1
                return True
            return False
        if title == SECTION_RESULT_Y:
            self.__sec_mode = 2
            return True
        return False

    def lines(self, lines: list[str]) -> bool:
        """Parse the lines."""
        if self.__sec_mode == 1:
            if self.__instance is not None:
                raise ValueError(
                    f"instance is already set to {self.__instance}.")
            key_1: Final[str] = "y.inst.name: "
            for line in lines:
                if line.startswith(key_1):
                    self.__instance = Instance.from_resource(
                        line[len(key_1):].strip())
            if self.__instance is None:
                raise ValueError(f"Did not find instance key {key_1!r} "
                                 f"in section {SECTION_SETUP}!")
        elif self.__sec_mode == 2:
            self.__packing_str = " ".join(lines).strip()
        else:
            raise ValueError("Should not be in section?")
        return (self.__instance is None) or (self.__packing_str is None)

    def end_file(self) -> bool:
        """End the file."""
        if self.__packing_str is None:
            raise ValueError(f"Section {SECTION_RESULT_Y} missing!")
        if self.__instance is None:
            raise ValueError(f"Section {SECTION_SETUP} missing or empty!")
        if self._result is not None:
            raise ValueError("Applied parser to more than one log file?")
        # pylint: disable=C0415,R0401
        from moptipyapps.binpacking2d.packing_space import (
            PackingSpace,  # pylint: disable=C0415,R0401
        )

        self._result = PackingSpace(self.__instance)\
            .from_str(self.__packing_str)
        self.__packing_str = None
        return False



"""An instance of the ordering problem."""


from math import isfinite
from typing import Callable, Final, Iterable, TypeVar, cast

import numpy as np
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.nputils import DEFAULT_INT
from moptipy.utils.strings import (
    num_to_str_for_name,
    sanitize_name,
)
from moptipy.utils.types import check_int_range, type_error
from scipy.stats import rankdata  # type: ignore

from moptipyapps.qap.instance import Instance as QAPInstance

#: the type variable for the source object
T = TypeVar("T")

#: the zero-based index column name
_ZERO_BASED_INDEX: Final[str] = "indexZeroBased"
#: the suggested x-coordinate
_SUGGESTED_X_IN_0_1: Final[str] = "suggestedXin01"


class Instance(QAPInstance):
    """
    An instance of the One-Dimensional Ordering Problem.

    Such an instance represents the ranking of objects by their distance to
    each other as a :mod:`~moptipyapps.qap` problem.

    >>> def _dist(a, b):
    ...     return abs(a - b)
    >>> def _tags(a):
    ...     return f"t{a}"
    >>> the_instance = Instance.from_sequence_and_distance(
    ...     [1, 2, 3], _dist, 1, 100, ("bla", ), _tags)
    >>> print(the_instance)
    order1d_3_1
    >>> print(the_instance.distances)
    [[0 1 2]
     [1 0 1]
     [2 1 0]]
    >>> print(the_instance.flows)
    [[0 4 2]
     [3 0 3]
     [2 4 0]]
    """

    def __init__(self, distances: np.ndarray,
                 flow_power: int | float,
                 horizon: int,
                 tag_titles: Iterable[str],
                 tags: Iterable[tuple[Iterable[str] | str, int]],
                 name: str | None = None) -> None:
        """
        Create an instance of the quadratic assignment problem.

        :param distances: the original distance matrix
        :param flow_power: the power to be used for constructing the flows
            based on the original distances
        :param horizon: the horizon for distance ranks, after which all
            elements are ignored
        :param tag_titles: the tag titles
        :param tags: the assignment of rows to names
        :param name: the instance name
        """
        if not isinstance(flow_power, int | float):
            raise type_error(flow_power, "flow_power", (int, float))
        if not (isfinite(flow_power) and (0 < flow_power < 100)):
            raise ValueError(
                f"flow_power must be in (0, 100), but is {flow_power}.")

        n: Final[int] = len(distances)  # the number of elements
        horizon = check_int_range(horizon, "horizon", 1, 1_000_000_000_000)

        # construct the distance matrix
        dist_matrix = np.zeros((n, n), DEFAULT_INT)
        for i in range(n):
            for j in range(i + 1, n):
                dist_matrix[i, j] = dist_matrix[j, i] = j - i

        # construct the flow matrix
        flows: np.ndarray = rankdata(
            distances, axis=1, method="average") - 1.0

        # get the flow rank multiplier
        multiplier: float = 1.0
        for i in range(n):
            for j in range(n):
                if i == j:
                    continue
                f = flows[i, j]
                if int(f) < f <= horizon:  # dow we need the multiplier?
                    multiplier = 0.5 ** (-flow_power)
                    break

        flow_matrix = np.zeros((n, n), DEFAULT_INT)
        max_val: Final[int] = min(n - 1, horizon)
        for i in range(n):
            for j in range(n):
                if i == j:
                    continue
                f = flows[i, j]
                if f > horizon:
                    continue
                flow_matrix[i, j] = int(round(
                    multiplier * ((max_val - f + 1) ** flow_power)))

        # construct a name if necessary
        if name is None:
            name = f"order1d_{n}_{num_to_str_for_name(flow_power)}"
            if max_val < (n - 1):
                name = f"{name}_{max_val}"

        super().__init__(dist_matrix, flow_matrix, name=name)

        if self.n != n:
            raise ValueError("error when assigning number of elements!")

        #: the flow power
        self.flow_power: Final[int | float] = flow_power
        #: the horizon
        self.horizon: Final[int] = max_val

        if not isinstance(tags, Iterable):
            raise type_error(tags, "tags", Iterable)
        if not isinstance(tag_titles, Iterable):
            raise type_error(tag_titles, "tag_titles", Iterable)

        #: the tag titles
        self.tag_titles: Final[tuple[str, ...]] = tuple(map(
            sanitize_name, tag_titles))
        req_len: Final[int] = len(self.tag_titles)
        if req_len <= 0:
            raise ValueError("No tags specified.")

        #: the tags, i.e., additional data for each original element
        self.tags: Final[tuple[tuple[tuple[str, ...], int], ...]] = (
            tuple(((t, ) if isinstance(t, str)
                   else tuple(t), k) for t, k in tags))
        if _ZERO_BASED_INDEX in self.tags:
            raise ValueError(f"{_ZERO_BASED_INDEX} not permitted in tags.")
        if _SUGGESTED_X_IN_0_1 in self.tags:
            raise ValueError(f"{_SUGGESTED_X_IN_0_1} not permitted in tags.")

        if len(self.tags) < n:
            raise ValueError(f"there must be at least {self.n} tags, but got "
                             f"{len(self.tags)}, i.e., {self.tags}")
        for tag in self.tags:
            check_int_range(tag[1], "id", 0, n)
            if len(tag[0]) != req_len:
                raise ValueError(
                    f"all tags must have the same length. "
                    f"while the first tag ({self.tags[0]}) has "
                    f"length {req_len}, {tag} has length {len(tag[0])}.")

    @staticmethod
    def from_sequence_and_distance(
            data: Iterable[T | None],
            get_distance: Callable[[T, T], int | float],
            flow_power: int | float,
            horizon: int,
            tag_titles: Iterable[str],
            get_tags: Callable[[T], str | Iterable[str]],
            name: str | None = None) -> "Instance":
        """
        Turn a sequence of objects into a One-Dimensional Ordering instance.

        :param data: the data source, i.e., an iterable of data elements
        :param get_tags: the function for extracting tags from objects
        :param get_distance: the function for getting the distance between
            objects
        :param flow_power: the flow power
        :param horizon: the maximal considered rank
        :param tag_titles: the tag names
        :param name: the optional name
        :return: the ordering instance

        >>> def _dist(a, b):
        ...     return abs(a - b)
        >>> def _tags(a):
        ...     return f"x{a}", f"b{a}"
        >>> res = Instance.from_sequence_and_distance(
        ...     [1, 2, 5], _dist, 2, 100, ("a", "b"), _tags)
        >>> print(res)
        order1d_3_2
        >>> print(res.flows)
        [[0 4 1]
         [4 0 1]
         [1 4 0]]
        >>> print(res.distances)
        [[0 1 2]
         [1 0 1]
         [2 1 0]]
        >>> res = Instance.from_sequence_and_distance(
        ...     [1, 2, 35, 4], _dist, 2, 100, ("a", "b"), _tags)
        >>> print(res)
        order1d_4_2
        >>> print(res.flows)
        [[0 9 1 4]
         [9 0 1 4]
         [1 4 0 9]
         [4 9 1 0]]
        >>> print(res.distances)
        [[0 1 2 3]
         [1 0 1 2]
         [2 1 0 1]
         [3 2 1 0]]
        >>> res = Instance.from_sequence_and_distance(
        ...     [1, 2, 4, 4], _dist, 2, 100, ("a", "b"), _tags)
        >>> print(res)
        order1d_3_2
        >>> print(res.flows)
        [[0 4 1]
         [4 0 1]
         [1 4 0]]
        >>> print(res.distances)
        [[0 1 2]
         [1 0 1]
         [2 1 0]]
        >>> print(res.tags)
        ((('x1', 'b1'), 0), (('x2', 'b2'), 1), (('x4', 'b4'), 2), \
(('x4', 'b4'), 2))
        >>> def _dist2(a, b):
        ...     return abs(abs(a) - abs(b)) + 1
        >>> res = Instance.from_sequence_and_distance(
        ...     [1, 2, -4, 4], _dist2, 2, 100, ("a", "b"), _tags)
        >>> print(res)
        order1d_4_2
        >>> print(res.flows)
        [[ 0 36  9  9]
         [36  0  9  9]
         [ 4 16  0 36]
         [ 4 16 36  0]]
        >>> print(res.distances)
        [[0 1 2 3]
         [1 0 1 2]
         [2 1 0 1]
         [3 2 1 0]]
        >>> res = Instance.from_sequence_and_distance(
        ...     [1, 2, -4, 4], _dist2, 3, 100, ("a", "b"), _tags)
        >>> print(res)
        order1d_4_3
        >>> print(res.flows)
        [[  0 216  27  27]
         [216   0  27  27]
         [  8  64   0 216]
         [  8  64 216   0]]
        >>> print(res.distances)
        [[0 1 2 3]
         [1 0 1 2]
         [2 1 0 1]
         [3 2 1 0]]
        >>> print(res.tags)
        ((('x1', 'b1'), 0), (('x2', 'b2'), 1), (('x-4', 'b-4'), 2), \
(('x4', 'b4'), 3))
        >>> res = Instance.from_sequence_and_distance(
        ...     [1, 2, -4, 4], _dist2, 3, 2, ("a", "b"), _tags)
        >>> print(res)
        order1d_4_3_2
        >>> print(res.flows)
        [[0 8 0 0]
         [8 0 0 0]
         [0 1 0 8]
         [0 1 8 0]]
        >>> print(res.distances)
        [[0 1 2 3]
         [1 0 1 2]
         [2 1 0 1]
         [3 2 1 0]]
        >>> print(res.tags)
        ((('x1', 'b1'), 0), (('x2', 'b2'), 1), (('x-4', 'b-4'), 2), \
(('x4', 'b4'), 3))
        >>> res = Instance.from_sequence_and_distance(
        ...     [1, 2, -4, 4], _dist2, 2, 2, ("a", "b"), _tags)
        >>> print(res)
        order1d_4_2_2
        >>> print(res.flows)
        [[0 4 0 0]
         [4 0 0 0]
         [0 1 0 4]
         [0 1 4 0]]
        >>> print(res.distances)
        [[0 1 2 3]
         [1 0 1 2]
         [2 1 0 1]
         [3 2 1 0]]
        """
        if not isinstance(data, Iterable):
            raise type_error(data, "data", Iterable)
        if not callable(get_tags):
            raise type_error(get_tags, "get_tags", call=True)
        if not callable(get_distance):
            raise type_error(get_distance, "get_distance", call=True)
        if not isinstance(tag_titles, Iterable):
            raise type_error(tag_titles, "tag_titles", Iterable)

        # build a distance matrix and purge all zero-distance elements
        datal: list[T] = cast(list[T], data) \
            if isinstance(data, list) else list(data)
        mappings: list[tuple[T, int]] = []
        distances: list[list[int | float]] = []
        i: int = 0
        while i < len(datal):
            o1: T = datal[i]
            j: int = i + 1
            current_dists: list[int | float] = [d[i] for d in distances]
            current_dists.append(0)
            while j < len(datal):
                o2: T = datal[j]
                dist: int | float = get_distance(o1, o2)
                if not (isfinite(dist) and (0 <= dist <= 1e100)):
                    raise ValueError(
                        f"invalid distance {dist} for objects {i}, {j}")
                if dist <= 0:  # distance == 0, must purge
                    mappings.append((o2, i))
                    del datal[j]
                    for ds in distances:
                        del ds[j]
                    continue
                current_dists.append(dist)
                j += 1
            mappings.append((o1, i))
            distances.append(current_dists)
            i += 1

        # we now got a full distance matrix, let's turn it into a rank matrix
        return Instance(np.array(distances),
                        flow_power, horizon, tag_titles,
                        ((get_tags(obj), idx) for obj, idx in mappings),
                        name)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("flowPower", self.flow_power)
        logger.key_value("horizon", self.horizon)
        logger.key_value("nOrig", len(self.tags))



"""
Examples of dynamic control problems.

Here we have examples for dynamic control problems. An
:mod:`~moptipyapps.dynamic_control.instance` of the dynamic control problem
is composed of

- a :mod:`~moptipyapps.dynamic_control.system` of differential equations that
  govern the state of a system and that also incorporate the output of a
  controller,
- a :mod:`~moptipyapps.dynamic_control.controller` blueprint, i.e., a function
  that can be parameterized and that computes a controller output from the
  system state, and
- an :mod:`~moptipyapps.dynamic_control.objective` that rates how well-behaved
  the system driven by the controller is in a simulation.

Such systems can serve as primitive testbeds of actual dynamic control
scenarios such as flight control or other fluid dynamic systems. Different
from such more complicated systems, they are much faster and easier to
simulate, though. So we can play with them much more easily and quickly.

The initial starting point of the work here were conversations with
Prof. Dr. Bernd NOACK and Guy Yoslan CORNEJO MACEDA of the Harbin Institute
of Technology in Shenzhen, China (哈尔滨工业大学(深圳)) as well as the
following two MSc theses and book:

1. Yuxiang LI (李宇翔). Jet Mixing Enhancement using Deep Reinforcement
   Learning (基于深度强化学习的射流混合增强控制). MSc Thesis. Harbin Institute
   of Technology in Shenzhen, China (哈尔滨工业大学(深圳)).
   January 2023.
2. Wei SUN (孙伟). Wake Control of 1-2-3 Fluidic Pinball using Deep
   Reinforcement Learning (基于深度强化学习方法的 1-2-3 流体弹球尾流控制).
   MSc Thesis. Harbin Institute of Technology in Shenzhen, China
   (哈尔滨工业大学(深圳)). January 2023.
3. Guy Yoslan CORNEJO MACEDA, François LUSSEYRAN, and Bernd R. NOACK.
   xMLC: A Toolkit for Machine Learning Control, First Edition.
   Machine Learning Tools in Fluid Mechanics, Vol 2.
   Shenzhen & Paris; Universitätsbibliothek der Technischen Universität
   Braunschweig. 2022 https://doi.org/10.24355/dbbs.084-202208220937-0
"""



"""
QAPLIB -- A Quadratic Assignment Problem Library.

1. QAPLIB - A Quadratic Assignment Problem Library. The Websites
   <https://qaplib.mgi.polymtl.ca/> (updated 2018) and
   <https://coral.ise.lehigh.edu/data-sets/qaplib/> (updated 2011), including
   the benchmark instances, on visited 2023-10-21.
2. Rainer E. Burkard, Stefan E. Karisch, and Franz Rendl. QAPLIB - A Quadratic
   Assignment Problem Library. Journal of Global Optimization. 10:391-403,
   1997. https://doi.org/10.1023/A:1008293323270.
"""

from importlib import resources  # nosem
from typing import TextIO


def open_resource_stream(file_name: str) -> TextIO:
    """
    Open a QAPLib resource stream.

    :param file_name: the file name of the resource
    :return: the stream
    """
    return resources.open_text(package=f"{__package__}", resource=file_name)



"""
An objective that counts constraint violations.

The idea is that we will probably not be able to always produce game plans
that adhere to all the constraints imposed by a Traveling Tournament Problem
:mod:`~moptipyapps.ttp` :mod:`~moptipyapps.ttp.instance`, so we will instead
probably usually generate game plans that may contain errors.

We will hope that optimization can take care of this by applying this
objective function here to get rid of them. In the documentation of function
:func:`~moptipyapps.ttp.errors.count_errors`, we explain the different types
of errors that may occur and that are counted.

This objective function plays thus well with encodings that produce infeasible
schedules, such as the very simple :mod:`~moptipyapps.ttp.game_encoding`.
"""


from typing import Final

import numba  # type: ignore
import numpy as np
from moptipy.api.objective import Objective
from moptipy.utils.nputils import int_range_to_dtype
from moptipy.utils.types import type_error

from moptipyapps.ttp.game_plan import GamePlan
from moptipyapps.ttp.instance import Instance


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def count_errors(y: np.ndarray, home_streak_min: int,
                 home_streak_max: int, away_streak_min: int,
                 away_streak_max: int, separation_min: int,
                 separation_max: int, temp_1: np.ndarray,
                 temp_2: np.ndarray) -> int:
    """
    Compute the number of errors in a game plan.

    This method counts the total number of the violations of any of the
    following constraints over `D = (n - 1) * rounds` days for `n`-team
    tournaments, where `rounds == 2` for double-round robin. The following
    kinds of errors are counted:

    1.  If a team `A` plays a team `B` on a given day, then `B` must play
        against `A` on that day and if `A` plays at home, then `B` must play
        away. If not, then that's an error. This can result in at most
        `D * n` errors, because each of the `n` teams has (at most) one game
        on each of the `D` days and if the *other* team could play against
        the wrong team (or does not play at all), then that's one error.
    2.  If a team has no other team assigned to play with, this is designated
        by value `0` and causes 1 error. This error also ends all ongoing
        streaks, i.e., may additionally lead to a streak violation of at
        most `max(home_streak_min, away_streak_min) - 1`. However, this
        cannot be more then two errors in sum per day (minus 1, for the
        first day). Also, this error is mutually exclusive with error 1.
        This can result in at most `D * n + (D - 1) * n = (2*D - 1) * n`
        errors. Since this error is also mutually exclusive with the errors
        from constraints 3 to 8 below, this gives us an upper bound of
        `(2*D - 1) * n` errors for all of the constraints (1-8) together.
    3.  A team has a home streak shorter than `home_streak_min`. No such error
        can occur on the first day. This error is mutually exclusive with
        error 2, as the streak violation is already counted there.
        This can result in at most `(D - 1) * n` errors, but this number is
        shared with the following constraints (4-6), because a streak can only
        either be a home streak or an away streak but not both and it can
        either be too short or too long, but not both.
    4.  A team has a home streak longer than `home_streak_max`. No such error
        can occur on the first day. This error is mutually exclusive with
        error 2.
    5.  A team has an away streak shorter than `away_streak_min`. No such
        error can occur on the first day. This error is mutually exclusive
        with error 2, as the streak violation is already counted there.
    6.  A team has an away streak longer than `away_streak_max`. No such
        error can occur on the first day. This error is mutually exclusive
        with error 2.
    7.  Team `A` plays team `B` *again* before the team has played at least
        `separation_min` games against other teams. This error cannot occur
        on the first day and is mutually exclusive with error 2.
        There can be most 1 such error per day for any pairing of teams and
        there are `n//2` pairings per day, giving us an upper bound of
        `D * n//2` errors in total. This error is mutually exclusive with
        the next constraint 8 (and constraint 2).
    8.  Team `A` plays team `B` *again* after the team has played more than
        `separation_max` games against other teams. This error cannot occur
        on the first day and is mutually exclusive with error 2.
        There can be most 1 such error per day for any pairing of teams and
        there are `n//2` pairings per day, giving us an upper bound of
        `D * n//2` errors in total. This error is mutually exclusive with
        the previous constraint 7 (and constraint 2).
    9.  If team `A` plays team `B` at home `a` times, then team `B` must play
        team `A` at home at least `a-1` and at most `a+1` times.
        In total, we have `D*n` games. There cannot be more than `(D*n) - 1`
        such errors. Notice that this kind of error can never occur if we use
        our :mod:`~moptipyapps.ttp.game_encoding` as representation.
    10. Each pairing of teams occurs as same as often, namely `rounds` times,
        with `rounds = 2` for double-round robin.
        In total, we have `D*n` games. There cannot be more than `D*n` such
        errors. Notice that this kind of error can never occur if we use
        our :mod:`~moptipyapps.ttp.game_encoding` as representation.

    The violations are counted on a per-day basis. For example, if
    `home_streak_max` is `3` and a team has a home streak of length `5`, then
    this counts as `2` errors. However, the errors are also counted in a
    non-redundant fashion: If a team `A` violates `separation_min` by, say,
    two days, then this counts as two errors. However, in this case, its
    opposing team `B` would have necessarily incured the exactly same
    violations. These are then not counted.

    As upper bound for the number of errors, we therefore have to add those of
    constraints 2, 9, and 10 and get `(2*D - 1) * n + D*n - 1 + D*n`, which
    gives us `(4*D - 1) * n - 1, where `D = (n - 1) * rounds`.
    The lower bound is obviously `0`.

    :param y: the game plan
    :param home_streak_min: the minimum permitted home streak length
    :param home_streak_max: the maximum permitted home streak length
    :param away_streak_min: the minimum permitted away streak length
    :param away_streak_max: the maximum permitted away streak length
    :param separation_min: the minimum number of games between a repetition
    :param separation_max: the maximum number games between a repetition
    :param temp_1: a temporary `n*(n-1)/2` integer array, which is used to
        hold, for each pairing, when the last game was played
    :param temp_2: a temporary `n,n` integer array, which is used to hold,
        how often each team played each other team
    :returns: `True` if a valid solution was produced

    >>> count_errors(np.array([[-2, 1], [2, -1]], int),
    ...              1, 3, 1, 3, 1, 2, np.empty(1, int),
    ...              np.empty((2, 2), int))
    1
    >>> count_errors(np.array([[2, -1], [-2, 1]], int),
    ...              1, 3, 1, 3, 1, 2, np.empty(1, int),
    ...              np.empty((2, 2), int))
    1
    >>> count_errors(np.array([[ 2, -1,  4, -3],
    ...                        [ 4,  3, -2, -1],
    ...                        [-2,  1, -4,  3],
    ...                        [-4, -3,  2,  1],
    ...                        [ 3,  4, -1, -2],
    ...                        [-3, -4,  1,  2]], int),
    ...              1, 3, 1, 3, 1, 2, np.empty(6, int),
    ...              np.empty((4, 4), int))
    2
    >>> count_errors(np.array([[ 2, -1,  4, -3],
    ...                        [ 4,  3, -2, -1],
    ...                        [-2,  1, -4,  3],
    ...                        [ 3,  4, -1, -2],
    ...                        [-4, -3,  2,  1],
    ...                        [-3, -4,  1,  2]], int),
    ...              1, 3, 1, 3, 1, 2, np.empty(6, int),
    ...              np.empty((4, 4), int))
    0
    >>> count_errors(np.array([[ 2, -1,  4, -3],
    ...                        [ 4,  3, -2, -1],
    ...                        [ 3,  4, -1, -2],
    ...                        [-2,  1, -4,  3],
    ...                        [-4, -3,  2,  1],
    ...                        [-3, -4,  1,  2]], int),
    ...              1, 3, 1, 3, 1, 2, np.empty(6, int),
    ...              np.empty((4, 4), int))
    0
    >>> count_errors(np.array([[ 2, -1,  4, -3],
    ...                        [ 4,  3, -2, -1],
    ...                        [ 3,  4, -1, -2],
    ...                        [-2,  1, -4,  3],
    ...                        [-4, -3,  2,  1],
    ...                        [-3, -4,  1,  2]], int),
    ...              1, 2, 1, 3, 1, 2, np.empty(6, int),
    ...              np.empty((4, 4), int))
    3
    >>> count_errors(np.array([[ 2, -1,  4, -3],
    ...                        [ 4,  3, -2, -1],
    ...                        [ 3,  4, -1, -2],
    ...                        [-2,  1, -4,  3],
    ...                        [-4, -3,  2,  1],
    ...                        [-3, -4,  1,  2]], int),
    ...              1, 2, 1, 2, 1, 2, np.empty(6, int),
    ...              np.empty((4, 4), int))
    6
    >>> count_errors(np.array([[ 2, -1,  4, -3],
    ...                        [ 4,  3, -2, -1],
    ...                        [ 3,  4, -1, -2],
    ...                        [-2,  1, -4,  3],
    ...                        [-4, -3,  2,  1],
    ...                        [-3, -4,  1,  2]], int),
    ...              1, 3, 1, 3, 1, 1, np.empty(6, int),
    ...              np.empty((4, 4), int))
    6
    """
    days, n = y.shape  # get the number of teams
    errors: int = 0  # the error counter
    temp_1.fill(-1)  # last time the teams played each other
    temp_2.fill(0)
    for team_1 in range(n):
        col = y[:, team_1]
        team_1_id: int = team_1 + 1
        is_in_home_streak: bool = False
        home_streak_len: int = -1
        is_in_away_streak: bool = False
        away_streak_len: int = -1

        for day, team_2_id in enumerate(col):
            if team_2_id == 0:  # is there no game on this day?
                errors += 1  # no game on this day == 1 error

                if is_in_away_streak:  # this ends away streaks
                    is_in_away_streak = False
                    if away_streak_len < away_streak_min:
                        errors += (away_streak_min - away_streak_len)
                    away_streak_len = -1
                elif is_in_home_streak:  # this ends home streaks, too
                    is_in_home_streak = False
                    if home_streak_len < home_streak_min:
                        errors += (home_streak_min - home_streak_len)
                    home_streak_len = -1
                continue  # nothing more to do here

            if team_2_id > 0:  # our team plays a home game

                # If team_2 > 0, this is a home game. So the other team
                # must have the corresponding index.
                team_2 = team_2_id - 1
                if y[day, team_2] != -team_1_id:
                    errors += 1
                # increase number of home games of team 1 against team 2
                temp_2[team_1, team_2] += 1

                if is_in_home_streak:  # if we are in a home streak...
                    home_streak_len += 1  # ...it continues
                    if home_streak_len > home_streak_max:
                        errors += 1  # too long? add to errors
                else:  # if we are not in home streak, it begins
                    is_in_home_streak = True
                    home_streak_len = 1
                    # if a home streak begins, any away streak ends
                    if is_in_away_streak:
                        if away_streak_len < away_streak_min:
                            errors += (away_streak_min - away_streak_len)
                        away_streak_len = -1
                        is_in_away_streak = False
            else:  # else: team_2 < 0 --> team_2 at home, team_1 away
                team_2 = (-team_2_id) - 1

                # This is an away game, so the other team must have the
                # corresponding id
                if y[day, team_2] != team_1_id:
                    errors += 1

                if is_in_away_streak:  # if we are in an away streak...
                    away_streak_len += 1  # ...the streak continues
                    if away_streak_len > away_streak_max:
                        errors += 1  # away streak too long? add to error
                else:  # team_1 away, but not in away streak
                    is_in_away_streak = True
                    away_streak_len = 1
                    if is_in_home_streak:
                        if home_streak_len < home_streak_min:
                            errors += (home_streak_min - home_streak_len)
                        is_in_home_streak = False
                        home_streak_len = 0

            # now we need to check for the game separation difference
            idx: int = ((team_1 * (team_1 - 1) // 2) + team_2) \
                if team_1 > team_2 \
                else ((team_2 * (team_2 - 1) // 2) + team_1)
            last_time: int = temp_1[idx]
            if last_time >= 0:
                if last_time < day:
                    difference = day - last_time - 1
                    if difference < separation_min:
                        errors += (separation_min - difference)
                    elif difference > separation_max:
                        errors += (difference - separation_max)
                else:
                    continue
            temp_1[idx] = day

    # sum up the team games
    games_per_combo: Final[int] = days // (n - 1)
    for i in range(n):
        for j in range(i):
            ij = temp_2[i, j]
            ji = temp_2[j, i]
            errors += abs(ij + ji - games_per_combo)
            diff = abs(ij - ji)
            if diff > 1:
                errors += diff - 1

    return errors


class Errors(Objective):
    """
    Compute the errors in a game plan.

    This objective function encompasses all the constraints imposed on
    standard TTP instances in one summarizing number. See the documentation
    of :func:`count_errors` for more information.
    """

    def __init__(self, instance: Instance) -> None:
        """
        Initialize the errors objective function.

        :param instance: the TTP instance
        """
        if not isinstance(instance, Instance):
            raise type_error(instance, "instance", Instance)
        super().__init__()

        #: the TTP instance
        self.instance: Final[Instance] = instance
        n: Final[int] = instance.n_cities
        # the data type for the temporary arrays
        dtype: Final[np.dtype] = int_range_to_dtype(
            -1, (n - 1) * instance.rounds)
        #: the internal temporary array 1
        self.__temp_1: Final[np.ndarray] = np.empty(n * (n - 1) // 2, dtype)
        #: the internal temporary array 2
        self.__temp_2: Final[np.ndarray] = np.empty((n, n), dtype)

    def evaluate(self, x: GamePlan) -> int:
        """
        Count the errors in a game plan as objective value.

        :param x: the game plan
        :return: the number of errors in the plan
        """
        inst: Final[Instance] = x.instance
        return count_errors(x, inst.home_streak_min, inst.home_streak_max,
                            inst.away_streak_min, inst.away_streak_max,
                            inst.separation_min, inst.separation_max,
                            self.__temp_1, self.__temp_2)

    def lower_bound(self) -> int:
        """
        Obtain the lower bound for errors: `0`, which means error-free.

        :return: `0`
        """
        return 0

    def upper_bound(self) -> int:
        """
        Compute upper bound for errors: `(4*D - 1) * n - 1`.

        Here `D` is the number of days, `n` is the number of teams, and
        `D = (n - 1) * rounds`. See the documentation of :func:`count_errors`.

        :return: `(4*D - 1) * n - 1`
        """
        n: Final[int] = self.instance.n_cities
        rounds: Final[int] = self.instance.rounds
        days: Final[int] = (n - 1) * rounds
        return (4 * days - 1) * n - 1

    def __str__(self) -> str:
        """Get the name of this objective function."""
        return "errors"



"""
Obtain Instances of the 2D-dimensional Bin Packing Problem.

With this program, we can obtain the instances of the two-dimensional bin
packing problem and convert them to a singular resource file.
The resource file holds one instance per line.

1. Manuel Iori, Vinícius Loti de Lima, Silvano Martello, and Michele Monaci.
   *2DPackLib*.
   https://site.unibo.it/operations-research/en/research/2dpacklib
2. Manuel Iori, Vinícius Loti de Lima, Silvano Martello, and Michele
   Monaci. 2DPackLib: A Two-Dimensional Cutting and Packing Library.
   *Optimization Letters* 16(2):471-480. March 2022.
   https://doi.org/10.1007/s11590-021-01808-y
"""

import zipfile
from io import BytesIO
from os.path import dirname, exists
from typing import Callable, Final, Iterable

# noinspection PyPackageRequirements
import certifi  # type: ignore

# noinspection PyPackageRequirements
import urllib3  # type: ignore
from moptipy.utils.path import Path
from moptipy.utils.temp import TempDir
from moptipy.utils.types import type_error

import moptipyapps.binpacking2d.instance as inst_mod
from moptipyapps.binpacking2d.instance import (
    INSTANCES_RESOURCE,
    Instance,
)

#: The base URLs of the relevant 2DPackLib instances.
__BASE_URLS: Final[Iterable[str]] = tuple(
    ("https://site.unibo.it/operations-research/en"
     f"/research/2dpacklib/{f}.zip") for f in ["a", "beng", "class"])


def download_2dpacklib_instances(
        dest_dir: str,
        source_urls: Iterable[str] = __BASE_URLS,
        http: urllib3.PoolManager = urllib3.PoolManager(
            cert_reqs="CERT_REQUIRED", ca_certs=certifi.where())) \
        -> Iterable[Path]:
    """
    Download the instances from 2DPackLib to a folder.

    This function downloads the instances from 2DPackLib, which are provided
    as zip archives containing one file per instance. It will extract all the
    instances into the folder `dest_dir` and return a tuple of the extracted
    files. You can specify the source URLs of the zip archives if you want,
    but by default we use the three instance sets `A`, `BENG`, and `CLASS`.

    :param dest_dir: the destination directory
    :param source_urls: the source URLs from which to download the zip
        archives with the 2DPackLib-formatted instances
    :param http: the HTTP pool
    :return: the list of unpackaged files
    """
    dest: Final[Path] = Path.directory(dest_dir)
    if not isinstance(source_urls, Iterable):
        raise type_error(source_urls, "source_urls", Iterable)
    result: Final[list[Path]] = []

    for i, url in enumerate(source_urls):  # iterate over the source URLs
        if not isinstance(url, str):
            raise type_error(url, f"source_urls[{i}]", str)

        response = http.request(  # download the zip archive
            "GET", url, redirect=True, retries=30)
        with zipfile.ZipFile(BytesIO(response.data), mode="r") as z:
            # unzip the ins2D files from the archives
            files: Iterable[str] = [f for f in z.namelist()
                                    if f.endswith(".ins2D")]
            paths: list[Path] = [dest.resolve_inside(f) for f in files]
            for path in paths:
                if exists(path):
                    raise ValueError(f"file {path} already exists!")
            z.extractall(dest, members=files)
            for path in paths:
                path.enforce_file()
                result.append(path)
    result.sort()
    return tuple(result)


def __normalize_2dpacklib_inst_name(instname: str) -> str:
    """
    Normalize an instance name.

    :param instname: the name
    :return: the normalized name
    """
    if not isinstance(instname, str):
        raise type_error(instname, "name", str)
    instname = instname.strip().lower()
    if (len(instname) == 2) and \
            (instname[0] in "abcdefghijklmnoprstuvwxyz") and \
            (instname[1] in "123456789"):
        return f"{instname[0]}0{instname[1]}"
    if instname.startswith("cl_"):
        return f"cl{instname[3:]}"
    return instname


def join_2dpacklib_instances_to_compact(
        files: Iterable[str], dest_file: str,
        normalizer: Callable[[str], str] = __normalize_2dpacklib_inst_name) \
        -> tuple[Path, Iterable[str]]:
    """
    Join all instances from a set of 2DPackLib files to one compact file.

    :param files: the iterable of 2DPackLib file paths
    :param dest_file: the destination file
    :param normalizer: the name normalizer, i.e., a function that processes
        and/or transforms an instance name
    :return: the canonical destination path and the list of instance names
        stored
    """
    if not isinstance(files, Iterable):
        raise type_error(files, "files", Iterable)
    if not callable(normalizer):
        raise type_error(normalizer, "normalizer", call=True)
    dest_path = Path.path(dest_file)
    data: Final[list[tuple[str, str]]] = []
    for f in files:
        inst: Instance = Instance.from_2dpacklib(Path.file(f))
        inst.name = normalizer(inst.name)
        data.append((inst.name, inst.to_compact_str()))
    data.sort()
    dest_path.write_all([content for _, content in data])
    dest_path.enforce_file()
    return dest_path, [thename for thename, _ in data]


def make_2dpacklib_resource(
        dest_file: str | None = None,
        source_urls: Iterable[str] = __BASE_URLS,
        normalizer: Callable[[str], str] = __normalize_2dpacklib_inst_name)\
        -> tuple[Path, Iterable[str]]:
    """
    Make the resource with all the relevant 2DPackLib instances.

    :param dest_file: the optional path to the destination file
    :param source_urls: the source URLs from which to download the zip
        archives with the 2DPackLib-formatted instances
    :param normalizer: the name normalizer, i.e., a function that processes
        and/or transforms an instance name
    :return: the canonical path to the and the list of instance names stored
    """
    dest_path: Final[Path] = Path.directory(dirname(inst_mod.__file__))\
        .resolve_inside(INSTANCES_RESOURCE) if dest_file is None \
        else Path.path(dest_file)
    with TempDir.create() as temp:
        files: Iterable[Path] = download_2dpacklib_instances(
            dest_dir=temp, source_urls=source_urls)
        return join_2dpacklib_instances_to_compact(
            files, dest_path, normalizer)


# create the tables if this is the main script
if __name__ == "__main__":
    _, names = make_2dpacklib_resource()
    rows: list[str] = ["_INSTANCES: Final[tuple[str, ...]] = ("]
    current = "    "
    has_space: bool = True
    for name in names:
        if (len(current) + (3 if has_space else 4) + len(name)) > 78:
            rows.append(current)
            current = "    "
            has_space = True
        current = f'{current}"{name}",' if has_space else \
            f'{current} "{name}",'
        has_space = False
    rows.append(current[:-1] + ")")

    print(  # noqa
        "#: the list of instance names of the 2DPackLib bin packing set")
    for s in rows:
        print(s)  # noqa



"""An extended end result statistics record to represent packings."""
import argparse
import os.path
from dataclasses import dataclass
from math import isfinite
from typing import Any, Callable, Final, Iterable, Mapping, cast

from moptipy.api.logging import (
    KEY_ALGORITHM,
    KEY_BEST_F,
    KEY_GOAL_F,
    KEY_INSTANCE,
    KEY_LAST_IMPROVEMENT_FE,
    KEY_LAST_IMPROVEMENT_TIME_MILLIS,
    KEY_MAX_FES,
    KEY_MAX_TIME_MILLIS,
    KEY_TOTAL_FES,
    KEY_TOTAL_TIME_MILLIS,
)
from moptipy.evaluation.base import (
    KEY_ENCODING,
    KEY_N,
    KEY_OBJECTIVE_FUNCTION,
    EvaluationDataElement,
)
from moptipy.evaluation.end_results import EndResult
from moptipy.evaluation.end_statistics import (
    KEY_BEST_F_SCALED,
    KEY_ERT_FES,
    KEY_ERT_TIME_MILLIS,
    KEY_N_SUCCESS,
    KEY_SUCCESS_FES,
    KEY_SUCCESS_TIME_MILLIS,
    EndStatistics,
)
from moptipy.evaluation.statistics import (
    EMPTY_CSV_ROW,
    Statistics,
)
from moptipy.utils.console import logger
from moptipy.utils.help import argparser
from moptipy.utils.logger import CSV_SEPARATOR
from moptipy.utils.path import Path
from moptipy.utils.strings import (
    num_to_str,
)
from moptipy.utils.types import check_int_range, immutable_mapping, type_error

from moptipyapps.binpacking2d.objectives.bin_count import BIN_COUNT_NAME
from moptipyapps.binpacking2d.packing_result import (
    _OBJECTIVE_LOWER,
    _OBJECTIVE_UPPER,
    KEY_BIN_HEIGHT,
    KEY_BIN_WIDTH,
    KEY_N_DIFFERENT_ITEMS,
    KEY_N_ITEMS,
    PackingResult,
)


@dataclass(frozen=True, init=False, order=False, eq=False)
class PackingStatistics(EvaluationDataElement):
    """
    An end statistics record of one run of one algorithm on one problem.

    This record provides the information of the outcome of one application of
    one algorithm to one problem instance in an immutable way.
    """

    #: the original end statistics record
    end_statistics: EndStatistics
    #: the number of items in the instance
    n_items: int
    #: the number of different items in the instance
    n_different_items: int
    #: the bin width
    bin_width: int
    #: the bin heigth
    bin_height: int
    #: the objective values evaluated after the optimization
    objectives: Mapping[str, Statistics]
    #: the bounds for the objective values (append ".lowerBound" and
    #: ".upperBound" to all objective function names)
    objective_bounds: Mapping[str, int]
    #: the bounds for the minimum number of bins of the instance
    bin_bounds: Mapping[str, int]

    def __init__(self,
                 end_statistics: EndStatistics,
                 n_items: int,
                 n_different_items: int,
                 bin_width: int,
                 bin_height: int,
                 objectives: Mapping[str, Statistics],
                 objective_bounds: Mapping[str, int | float],
                 bin_bounds: Mapping[str, int]):
        """
        Create a consistent instance of :class:`PackingStatistics`.

        :param end_statistics: the end statistics
        :param n_items: the number of items
        :param n_different_items: the number of different items
        :param bin_width: the bin width
        :param bin_height: the bin height
        :param objectives: the objective values computed after the
            optimization
        :param bin_bounds: the different bounds for the number of bins
        :param objective_bounds: the bounds for the objective functions
        :raises TypeError: if any parameter has a wrong type
        :raises ValueError: if the parameter values are inconsistent
        """
        super().__init__()
        if not isinstance(end_statistics, EndStatistics):
            raise type_error(end_statistics, "end_statistics", EndResult)
        if end_statistics.best_f != objectives[end_statistics.objective]:
            raise ValueError(
                f"end_statistics.best_f={end_statistics.best_f}, but "
                f"objectives[{end_statistics.objective!r}]="
                f"{objectives[end_statistics.objective]}.")
        if not isinstance(objectives, Mapping):
            raise type_error(objectives, "objectives", Mapping)
        if not isinstance(objective_bounds, Mapping):
            raise type_error(objective_bounds, "objective_bounds", Mapping)
        if not isinstance(bin_bounds, Mapping):
            raise type_error(bin_bounds, "bin_bounds", Mapping)
        if len(objective_bounds) != (2 * len(objectives)):
            raise ValueError(f"it is required that there is a lower and an "
                             f"upper bound for each of the {len(objectives)} "
                             f"functions, but we got {len(objective_bounds)} "
                             f"bounds, objectives={objectives}, "
                             f"objective_bounds={objective_bounds}.")

        for name, value in objectives.items():
            if not isinstance(name, str):
                raise type_error(
                    name, f"name of evaluation[{name!r}]={value!r}", str)
            if not isinstance(value, Statistics):
                raise type_error(
                    value, f"value of evaluation[{name!r}]={value!r}",
                    Statistics)
            lower = objective_bounds[f"{name}{_OBJECTIVE_LOWER}"]
            if not isfinite(lower):
                raise ValueError(f"{name}{_OBJECTIVE_LOWER}=={lower}.")
            upper = objective_bounds[f"{name}{_OBJECTIVE_UPPER}"]
            if not (lower <= value.minimum <= value.maximum <= upper):
                raise ValueError(
                    f"it is required that {name}{_OBJECTIVE_LOWER}<={name}."
                    f"min<={name}.max<={name}{_OBJECTIVE_UPPER}, but got "
                    f"{lower}, {value.minimum}, {value.maximum} and {upper}.")

        bins: Final[Statistics | None] = cast(
            Statistics, objectives[BIN_COUNT_NAME]) \
            if BIN_COUNT_NAME in objectives else None
        for name, value2 in bin_bounds.items():
            if not isinstance(name, str):
                raise type_error(
                    name, f"name of bounds[{name!r}]={value2!r}", str)
            check_int_range(value2, f"bounds[{name!r}]", 1, 1_000_000_000)
            if (bins is not None) and (bins.minimum < value2):
                raise ValueError(
                    f"number of bins={bins} is inconsistent with "
                    f"bound {name!r}={value2}.")

        object.__setattr__(self, "end_statistics", end_statistics)
        object.__setattr__(self, "objectives", immutable_mapping(objectives))
        object.__setattr__(self, "objective_bounds",
                           immutable_mapping(objective_bounds))
        object.__setattr__(self, "bin_bounds", immutable_mapping(bin_bounds))
        object.__setattr__(self, "n_different_items", check_int_range(
            n_different_items, "n_different_items", 1, 1_000_000_000_000))
        object.__setattr__(self, "n_items", check_int_range(
            n_items, "n_items", n_different_items, 1_000_000_000_000))
        object.__setattr__(self, "bin_width", check_int_range(
            bin_width, "bin_width", 1, 1_000_000_000_000))
        object.__setattr__(self, "bin_height", check_int_range(
            bin_height, "bin_height", 1, 1_000_000_000_000))

    def _tuple(self) -> tuple[Any, ...]:
        """
        Create a tuple with all the data of this data class for comparison.

        :returns: a tuple with all the data of this class, where `None` values
            are masked out
        """
        # noinspection PyProtectedMember
        return self.end_statistics._tuple()

    @staticmethod
    def from_packing_results(
            results: Iterable[PackingResult],
            collector: Callable[["PackingStatistics"], None]) -> None:
        """
        Create packing statistics from a sequence of packing results.

        :param results: the packing results
        :param collector: the collector receiving the created packing
            statistics
        """
        if not isinstance(results, Iterable):
            raise type_error(results, "results", Iterable)
        if not callable(collector):
            raise type_error(collector, "collector", call=True)
        groups: Final[dict[tuple[str, str, str, str], list[PackingResult]]] \
            = {}
        objectives_set: set[str] = set()
        for i, pr in enumerate(results):
            if not isinstance(pr, PackingResult):
                raise type_error(pr, f"end_results[{i}]", PackingResult)
            setting: tuple[str, str, str, str] = \
                (pr.end_result.algorithm, pr.end_result.instance,
                 pr.end_result.objective, "" if pr.end_result.encoding is None
                 else pr.end_result.encoding)
            if setting in groups:
                groups[setting].append(pr)
            else:
                groups[setting] = [pr]
            objectives_set.update(pr.objectives.keys())

        if len(groups) <= 0:
            raise ValueError("results is empty!")
        if len(objectives_set) <= 0:
            raise ValueError("results has not objectives!")
        end_stats: Final[list[EndStatistics]] = []
        objectives: Final[list[str]] = sorted(objectives_set)

        for key in sorted(groups.keys()):
            data = groups[key]

            pr0 = data[0]
            n_items: int = pr0.n_items
            n_different_items: int = pr0.n_different_items
            bin_width: int = pr0.bin_width
            bin_height: int = pr0.bin_height
            used_objective: str = pr0.end_result.objective
            encoding: str | None = pr0.end_result.encoding
            if used_objective not in objectives_set:
                raise ValueError(
                    f"{used_objective!r} not in {objectives_set!r}.")
            if used_objective != key[2]:
                raise ValueError(
                    f"used objective={used_objective!r} different "
                    f"from key[2]={key[2]}!?")
            if (encoding is not None) and (encoding != key[3]):
                raise ValueError(
                    f"used encoding={encoding!r} different "
                    f"from key[3]={key[3]}!?")
            objective_bounds: Mapping[str, int | float] = pr0.objective_bounds
            bin_bounds: Mapping[str, int] = pr0.bin_bounds
            for i, pr in enumerate(data):
                if n_items != pr.n_items:
                    raise ValueError(f"n_items={n_items} for data[0] but "
                                     f"{pr.n_items} for data[{i}]?")
                if n_different_items != pr.n_different_items:
                    raise ValueError(
                        f"n_different_items={n_different_items} for data[0] "
                        f"but {pr.n_different_items} for data[{i}]?")
                if bin_width != pr.bin_width:
                    raise ValueError(
                        f"bin_width={bin_width} for data[0] "
                        f"but {pr.bin_width} for data[{i}]?")
                if bin_height != pr.bin_height:
                    raise ValueError(
                        f"bin_height={bin_height} for data[0] "
                        f"but {pr.bin_height} for data[{i}]?")
                if used_objective != pr.end_result.objective:
                    raise ValueError(
                        f"used objective={used_objective!r} for data[0] "
                        f"but {pr.end_result.objective!r} for data[{i}]?")
                if objective_bounds != pr.objective_bounds:
                    raise ValueError(
                        f"objective_bounds={objective_bounds!r} for data[0] "
                        f"but {pr.objective_bounds!r} for data[{i}]?")
                if bin_bounds != pr.bin_bounds:
                    raise ValueError(
                        f"bin_bounds={bin_bounds!r} for data[0] "
                        f"but {pr.bin_bounds!r} for data[{i}]?")

            EndStatistics.from_end_results((pr.end_result for pr in data),
                                           end_stats.append)
            if len(end_stats) != 1:
                raise ValueError(f"got {end_stats} from {data}?")

            collector(PackingStatistics(
                end_statistics=end_stats[0],
                n_items=n_items,
                n_different_items=n_different_items,
                bin_width=bin_width,
                bin_height=bin_height,
                objectives={
                    o: Statistics.create([pr.objectives[o] for pr in data])
                    for o in objectives
                },
                objective_bounds=objective_bounds,
                bin_bounds=bin_bounds,
            ))
            end_stats.clear()

    @staticmethod
    def to_csv(results: Iterable["PackingStatistics"], file: str) -> Path:
        """
        Write a sequence of packing statistics to a file in CSV format.

        :param results: the end statistics
        :param file: the path
        :return: the path of the file that was written
        """
        path: Final[Path] = Path.path(file)
        logger(f"Writing packing results to CSV file {path!r}.")
        Path.path(os.path.dirname(path)).ensure_dir_exists()

        # get a nicely sorted view on the statistics
        use_stats = sorted(results)

        has_algorithm: bool = False  # 1
        has_instance: bool = False  # 2
        has_objective: bool = False  # 4
        has_encoding: bool = False  # 8
        has_goal_f: int = 0  # 16
        has_best_f_scaled: bool = False  # 32
        has_n_success: bool = False  # 64
        has_success_fes: bool = False  # 128
        has_success_time_millis: bool = False  # 256
        has_ert_fes: bool = False  # 512
        has_ert_time_millis: bool = False  # 1024
        has_max_fes: int = 0  # 2048
        has_max_time_millis: int = 0  # 4096
        checker: int = 8191

        for ess in use_stats:
            es = ess.end_statistics
            if es.algorithm is not None:
                has_algorithm = True
                checker &= ~1
            if es.instance is not None:
                has_instance = True
                checker &= ~2
            if es.objective is not None:
                has_objective = True
                checker &= ~4
            if es.encoding is not None:
                has_encoding = True
                checker &= ~8
            if es.goal_f is not None:
                if isinstance(es.goal_f, Statistics) \
                        and (es.goal_f.maximum > es.goal_f.minimum):
                    has_goal_f = 2
                    checker &= ~16
                elif has_goal_f == 0:
                    has_goal_f = 1
            if es.best_f_scaled is not None:
                has_best_f_scaled = True
                checker &= ~32
            if es.n_success is not None:
                has_n_success = True
                checker &= ~64
            if es.success_fes is not None:
                has_success_fes = True
                checker &= ~128
            if es.success_time_millis is not None:
                has_success_time_millis = True
                checker &= ~256
            if es.ert_fes is not None:
                has_ert_fes = True
                checker &= ~512
            if es.ert_time_millis is not None:
                has_ert_time_millis = True
                checker &= ~1024
            if es.max_fes is not None:
                if isinstance(es.max_fes, Statistics) \
                        and (es.max_fes.maximum > es.max_fes.minimum):
                    has_max_fes = 2
                    checker &= ~2048
                elif has_max_fes == 0:
                    has_max_fes = 1
            if es.max_time_millis is not None:
                if isinstance(es.max_time_millis, Statistics) \
                        and (es.max_time_millis.maximum
                             > es.max_time_millis.minimum):
                    has_max_time_millis = 2
                    checker &= ~4096
                elif has_max_time_millis == 0:
                    has_max_time_millis = 1
            if checker == 0:
                break

        # get the names of the bounds and objectives
        bin_bounds_set: set[str] = set()
        objectives_set: set[str] = set()
        for pr in use_stats:
            bin_bounds_set.update(pr.bin_bounds.keys())
            objectives_set.update(pr.objectives.keys())
        bin_bounds: Final[list[str]] = sorted(bin_bounds_set)
        objectives: Final[list[str]] = sorted(objectives_set)

        with path.open_for_write() as out:
            wrt: Final[Callable] = out.write
            sep: Final[str] = CSV_SEPARATOR
            if has_algorithm:
                wrt(KEY_ALGORITHM)
                wrt(sep)
            if has_instance:
                wrt(KEY_INSTANCE)
                wrt(sep)
            if has_objective:
                wrt(KEY_OBJECTIVE_FUNCTION)
                wrt(sep)
            if has_encoding:
                wrt(KEY_ENCODING)
                wrt(sep)
            wrt(KEY_N_ITEMS)
            wrt(sep)
            wrt(KEY_N_DIFFERENT_ITEMS)
            wrt(sep)
            wrt(KEY_BIN_WIDTH)
            wrt(sep)
            wrt(KEY_BIN_HEIGHT)
            wrt(sep)
            for bb in bin_bounds:
                wrt(bb)
                wrt(sep)
            for oo in objectives:
                wrt(oo)
                wrt(_OBJECTIVE_LOWER)
                wrt(sep)
                wrt(oo)
                wrt(_OBJECTIVE_UPPER)
                wrt(sep)

            def h(p) -> None:
                wrt(sep.join(Statistics.csv_col_names(p)))

            wrt(KEY_N)
            wrt(sep)
            h(KEY_BEST_F)
            wrt(sep)
            h(KEY_LAST_IMPROVEMENT_FE)
            wrt(sep)
            h(KEY_LAST_IMPROVEMENT_TIME_MILLIS)
            wrt(sep)
            h(KEY_TOTAL_FES)
            wrt(sep)
            h(KEY_TOTAL_TIME_MILLIS)
            if has_goal_f == 1:
                wrt(sep)
                wrt(KEY_GOAL_F)
            elif has_goal_f == 2:
                wrt(sep)
                h(KEY_GOAL_F)
            if has_best_f_scaled:
                wrt(sep)
                h(KEY_BEST_F_SCALED)
            if has_n_success:
                wrt(sep)
                wrt(KEY_N_SUCCESS)
            if has_success_fes:
                wrt(sep)
                h(KEY_SUCCESS_FES)
            if has_success_time_millis:
                wrt(sep)
                h(KEY_SUCCESS_TIME_MILLIS)
            if has_ert_fes:
                wrt(sep)
                wrt(KEY_ERT_FES)
            if has_ert_time_millis:
                wrt(sep)
                wrt(KEY_ERT_TIME_MILLIS)
            if has_max_fes == 1:
                wrt(sep)
                wrt(KEY_MAX_FES)
            elif has_max_fes == 2:
                wrt(sep)
                h(KEY_MAX_FES)
            if has_max_time_millis == 1:
                wrt(sep)
                wrt(KEY_MAX_TIME_MILLIS)
            elif has_max_time_millis == 2:
                wrt(sep)
                h(KEY_MAX_TIME_MILLIS)
            for oo in objectives:
                wrt(sep)
                h(oo)

            out.write("\n")

            csv: Final[Callable] = Statistics.value_to_csv
            num: Final[Callable] = num_to_str

            for rec in use_stats:
                er = rec.end_statistics
                if has_algorithm:
                    if er.algorithm is not None:
                        wrt(er.algorithm)
                    wrt(sep)
                if has_instance:
                    if er.instance is not None:
                        wrt(er.instance)
                    wrt(sep)
                if has_objective:
                    if er.objective is not None:
                        wrt(er.objective)
                    wrt(sep)
                if has_encoding:
                    if er.encoding is not None:
                        wrt(er.encoding)
                    wrt(sep)
                wrt(str(rec.n_items))
                wrt(sep)
                wrt(str(rec.n_different_items))
                wrt(sep)
                wrt(str(rec.bin_width))
                wrt(sep)
                wrt(str(rec.bin_height))
                wrt(sep)

                for bb in bin_bounds:
                    wrt(str(rec.bin_bounds[bb]))
                    wrt(sep)
                for oo in objectives:
                    wrt(num_to_str(rec.objective_bounds[
                        f"{oo}{_OBJECTIVE_LOWER}"]))
                    wrt(sep)
                    wrt(num_to_str(rec.objective_bounds[
                        f"{oo}{_OBJECTIVE_UPPER}"]))
                    wrt(sep)

                wrt(str(er.n))
                wrt(sep)
                wrt(er.best_f.to_csv())
                wrt(sep)
                wrt(er.last_improvement_fe.to_csv())
                wrt(sep)
                wrt(er.last_improvement_time_millis.to_csv())
                wrt(sep)
                wrt(er.total_fes.to_csv())
                wrt(sep)
                wrt(er.total_time_millis.to_csv())
                if has_goal_f == 1:
                    wrt(sep)
                    if er.goal_f is not None:
                        wrt(num(er.goal_f.median if isinstance(
                            er.goal_f, Statistics) else er.goal_f))
                elif has_goal_f == 2:
                    wrt(sep)
                    if isinstance(er.goal_f, Statistics):
                        wrt(er.goal_f.to_csv())
                    elif isinstance(er.goal_f, int | float):
                        wrt(csv(er.goal_f))
                    else:
                        wrt(EMPTY_CSV_ROW)
                if has_best_f_scaled:
                    wrt(sep)
                    if er.best_f_scaled is None:
                        wrt(EMPTY_CSV_ROW)
                    else:
                        wrt(er.best_f_scaled.to_csv())
                if has_n_success:
                    wrt(sep)
                    if er.n_success is not None:
                        wrt(str(er.n_success))
                if has_success_fes:
                    wrt(sep)
                    if er.success_fes is None:
                        wrt(EMPTY_CSV_ROW)
                    else:
                        wrt(er.success_fes.to_csv())
                if has_success_time_millis:
                    wrt(sep)
                    if er.success_time_millis is None:
                        wrt(EMPTY_CSV_ROW)
                    else:
                        wrt(er.success_time_millis.to_csv())
                if has_ert_fes:
                    wrt(sep)
                    if er.ert_fes is not None:
                        wrt(num(er.ert_fes))
                if has_ert_time_millis:
                    wrt(sep)
                    if er.ert_time_millis is not None:
                        wrt(num(er.ert_time_millis))
                if has_max_fes == 1:
                    wrt(sep)
                    if er.max_fes is not None:
                        wrt(str(er.max_fes.median if isinstance(
                            er.max_fes, Statistics) else er.max_fes))
                elif has_max_fes == 2:
                    wrt(sep)
                    if isinstance(er.max_fes, Statistics):
                        wrt(er.max_fes.to_csv())
                    elif isinstance(er.max_fes, int | float):
                        wrt(csv(er.max_fes))
                    else:
                        wrt(EMPTY_CSV_ROW)
                if has_max_time_millis == 1:
                    wrt(sep)
                    if er.max_time_millis is not None:
                        wrt(str(er.max_time_millis.median if isinstance(
                            er.max_time_millis, Statistics)
                            else er.max_time_millis))
                elif has_max_time_millis == 2:
                    wrt(sep)
                    if isinstance(er.max_time_millis, Statistics):
                        wrt(er.max_time_millis.to_csv())
                    elif isinstance(er.max_time_millis, int | float):
                        wrt(csv(er.max_time_millis))
                    else:
                        wrt(EMPTY_CSV_ROW)
                for oo in objectives:
                    wrt(sep)
                    wrt(rec.objectives[oo].to_csv())
                out.write("\n")

        # finally, return the path to the generated file
        return path


# Run packing-results to stat file if executed as script
if __name__ == "__main__":
    parser: Final[argparse.ArgumentParser] = argparser(
        __file__, "Build an end-results statistics CSV file.",
        "This program computes statistics over packing results")
    def_src: str = "./evaluation/end_results.txt"
    if not os.path.isfile(def_src):
        def_src = "./results"
    parser.add_argument(
        "source", nargs="?", default=def_src,
        help="either the directory with moptipy log files or the path to the "
             "end-results CSV file", type=Path.path)
    parser.add_argument(
        "dest", type=Path.path, nargs="?",
        default="./evaluation/end_statistics.txt",
        help="the path to the end results statistics CSV file to be created")
    args: Final[argparse.Namespace] = parser.parse_args()

    src_path: Final[Path] = args.source
    packing_results: Final[list[PackingResult]] = []
    if src_path.is_file():
        logger(f"{src_path!r} identifies as file, load as end-results csv")
        PackingResult.from_csv(src_path, packing_results.append)
    else:
        logger(f"{src_path!r} identifies as directory, load it as log files")
        PackingResult.from_logs(src_path, packing_results.append)

    packing_stats: Final[list[PackingStatistics]] = []
    PackingStatistics.from_packing_results(
        results=packing_results, collector=packing_stats.append)
    PackingStatistics.to_csv(packing_stats, args.dest)



"""
Codes for the two-dimensional bin packing problem.

- :mod:`~moptipyapps.binpacking2d.instance` provides the instance data of
  two-dimensional bin packing problems.
  The instance data comprises the object sizes and multiplicities as well as
  the bin sizes.
  Several default instances from \
[2DPackLib](https://site.unibo.it/operations-research/en/research/2dpacklib)
  can be loaded from resources.
- :mod:`~moptipyapps.binpacking2d.packing` can store a packing, i.e., an
  assignment of objects to coordinates and bins.
  Such packings are solutions to the bin packing problem.
- :mod:`~moptipyapps.binpacking2d.packing_space` provides an implementation of
  the :mod:`~moptipy.api.space` interface for the
  :mod:`~moptipyapps.binpacking2d.packing` objects.
  This class allows, for example, to instantiate the packings, to verify
  whether they are correct, and to convert them to and from strings.
- :mod:`~moptipyapps.binpacking2d.plot_packing` allows you to plot a packing.
- :mod:`~moptipyapps.binpacking2d.encodings.ibl_encoding_1` is an
  implementation of the improved bottom-left encoding which closes bins
  immediately once an object does not fit.
- :mod:`~moptipyapps.binpacking2d.encodings.ibl_encoding_2` is another
  implementation of the improved bottom-left encoding which tests each bin for
  each object.
- :mod:`~moptipyapps.binpacking2d.objectives.bin_count_and_last_empty`
  provides an objective function that tries to minimize the number of bins and
  pushes towards decreasing the number of objects in the very last bin used.
- :mod:`~moptipyapps.binpacking2d.make_instances` offers a method to download
  the two-dimensional bin packing instances from the original sources in the \
[2DPackLib](https://site.unibo.it/operations-research/en/research/2dpacklib)
  format.

Important initial work on this code has been contributed by Mr. Rui ZHAO
(赵睿), <zr1329142665@163.com> a Master's student at the Institute of Applied
Optimization (应用优化研究所, http://iao.hfuu.edu.cn) of the School of
Artificial Intelligence and Big Data (人工智能与大数据学院) at Hefei University
(合肥学院) in Hefei, Anhui, China (中国安徽省合肥市) under the supervision of
Prof. Dr. Thomas Weise (汤卫思教授).
"""
